# encoding: utf-8

from __future__ import print_function
import time


class SimpleTimer(object):
    """All time are counted relative to start."""
    def __init__(self):
        self.start = 0
        self.events = []
    
    def tick(self, event_name="Untitled"):
        now = time.time()
        if not self.start:
            self.start = now
            last = now
        else:
            _, last = self.events[-1]
        total = now - self.start
        delta = now - last
        
        self.events.append((event_name, now))
        
        print("Total {:3.1f}    | Delta {:3.1f}    | {:s}".format(total, delta, event_name))



_prof_data = {}

class ProfData :
    def __init__(self, name):
        self.name = name
        self.begin_time = time.time()
        self.used_time = 0

    def end(self):
        self.end_time = time.time()
        self.used_time =  self.end_time - self.begin_time



def prof_sample_begin(name):
    data = ProfData(name)
    if name in _prof_data:
        _prof_data[name].append(data)
    else:
        _prof_data[name] = [data]
    return data

def prof_sample_end(data):
    data.end()

def prof_sample(name, func):
    d = prof_sample_begin(name)
    v = func()
    d.end()
    return v

def prof_print():

    print_data = []
    for data in _prof_data.values():
        name = data[0].name
        max_time = 0.0
        min_time = 10000000000
        total_time = 0
        for d in data:
            total_time += d.used_time
            if d.used_time > max_time: max_time = d.used_time
            if d.used_time < min_time: min_time = d.used_time

        print_data.append( { 'name' : name,
                             'max': max_time,
                             'min': min_time,
                             'total': total_time,
                             'count':len(data),
                             'avg': total_time / len(data)
                             })

    print_data = sorted(print_data, key=lambda x : -x['total'])

    for d in print_data:
        print ("prof: {:<30} count {:8d} total {:8.4f}s, min {:8.4f}s, max {:8.4f}s, avg {:8.4f}"
               .format(d['name'], d['count'], d['total'], d['min'], d['max'], d['avg']))





import numpy as np
import pandas as pd
import datetime
from jaqs.data import RemoteDataService
import re

# -------------------------------------------------------------------
# Settings
# -------------------------------------------------------------------
dataview_dir_path = '../../output/continue_contract'

ds = RemoteDataService()
ds.init_from_config(data_config)


def get_symbol_date_map(data_service, df_inst_, start_date, end_date, days_to_delist):
    """
    Get a map {trade_date -> front month contract symbol}

    """
    # get the trade date list between start_date and end_date
    dates = data_service.query_trade_dates(start_date, end_date)
    symbols_list = []
    
    mask_within_range = (df_inst_['delist_date'] > dates[0]) & (df_inst_['list_date'] < dates[-1])
    df_inst_ = df_inst_.loc[mask_within_range]
    
    j = 0
    for i, td in enumerate(dates):
        delist_date = df_inst_['delist_date'].iat[j]
        idx = np.nonzero(dates == delist_date)
        
        if (delist_date <= dates[-1]) and (idx[0][0] - i <= days_to_delist):
            j += 1
            delist_date = df_inst_['delist_date'].iat[j]
        symbol = df_inst_['symbol'].iat[j]
        symbols_list.append(symbol)
    
    res = pd.DataFrame(data={'trade_date': dates, 'symbol': symbols_list})
    res.loc[:, 'symbol_next'] = res['symbol'].shift(-1).fillna(method='ffill')
    return res


def get_continuous_contract(symbol, start_date, end_date, change_date, fields):
    # start_date = ds.get_last_trade_date(start_date)
    
    # get information of all instruments and filter only relevant contracts
    df_inst = ds.query_inst_info(symbol="", fields=','.join(['symbol', 'inst_type',
                                                             'market', 'status',
                                                             'multiplier', 'list_date',
                                                             'delist_date']))
    df_inst = df_inst.reset_index()
    def is_target_symbol(s):
        return len(re.findall(symbol + r'\d+', s)) > 0
    mask_stock_index_future = df_inst['symbol'].apply(is_target_symbol)
    df_inst = df_inst.loc[mask_stock_index_future].sort_values('delist_date')
    # df_inst.index = range(len(df_inst))

    # get the front month contract for each trading date
    # prevent the case that no contracts exist near start_date
    first_list_date = df_inst['list_date'].min()
    if start_date < first_list_date:
        start_date = first_list_date

    df_inst_map = get_symbol_date_map(ds, df_inst, start_date, end_date, change_date)
    # df_inst_map['pre_trade_date'] = df_inst_map['trade_date'].shift(1)
    # df_inst_map = df_inst_map.dropna()
    # df_inst_map['pre_trade_date'] = df_inst_map['pre_trade_date'].astype(np.integer)

    # get the daily info
    df_future_daily = query_daily_data(ds, df_inst_map, fields)

    df_future_daily = df_future_daily.drop(['freq'], axis=1)
    '''
    df_future_daily['move'] = np.where(df_future_daily['symbol'] != df_future_daily['symbol'].shift(-1).fillna(method='ffill'), 1, 0)
    # df_future_daily['ret_shift'] = 0.0

    #df_future_daily.index = range(len(df_future_daily))

    for i in range(len(df_future_daily) - 1):
        if df_future_daily.ix[i, 'move'] == 1.0:
            df_future_daily.ix[i, 'ret_shift'] = np.log(df_future_daily.ix[i+1, 'close']) - np.log(df_future_daily.ix[i, 'close'])

    df_future_daily.index = df_future_daily['trade_date']
    df_future_daily['trade_date'] = df_future_daily['trade_date'].apply(lambda x: datetime.datetime.strptime(str(x), '%Y%m%d').date())
    '''
    return df_future_daily


def query_daily_data(data_service, df_inst_map_, fields_):
    l = []
    
    for symbol in df_inst_map_['symbol'].unique():
        df_symbol = df_inst_map_.loc[df_inst_map_['symbol'] == symbol]
        start_date, end_date = df_symbol['trade_date'].iat[0], df_symbol['trade_date'].iat[-1]
        symbol_next = df_symbol['symbol_next'].iat[-1]
        
        df, _ = data_service.daily(symbol, start_date=start_date, end_date=end_date, fields=fields_)
        df_next, _ = data_service.daily(symbol_next, start_date=end_date, end_date=end_date, fields=fields_)
        
        close_diff = df_next['close'].iat[-1] - df['close'].iat[-1]
        for col in ['open', 'high', 'low', 'close']:
            df.loc[:, col] = df[col] + close_diff
            
        df.loc[:, 'symbol_next'] = df['symbol']
        df.loc[:, 'symbol_next'].iat[-1] = symbol_next
        df.loc[:, 'close_diff'] = 0.0
        df.loc[:, 'close_diff'].iat[-1] = close_diff
        # df_future['pre_close'] = df_future['close'].shift(1)
        # df_future['log_ret'] = np.log(df_future['close']).diff()
        # df_future = df_future.dropna()
        l.append(df)
    
    res = pd.concat(l, axis=0)
    return res


if __name__ == '__main__':
    future_df = get_continuous_contract('rb', 20170101, 20171225, 3, 'open,high,low,close')


# encoding: utf-8

import numpy as np
import pandas as pd

from jaqs.util import numeric


def to_quantile(df, n_quantiles=5, axis=1):
    """
    Convert cross-section values to the quantile number they belong.
    Small values get small quantile numbers.
    
    Parameters
    ----------
    df : pd.DataFrame or pd.Series
        index date, column symbols
    n_quantiles : int
        The number of quantile to be divided to.
    axis : int
        Axis to apply quantilize.

    Returns
    -------
    res : DataFrame
        index date, column symbols

    """
    # TODO: unnecesssary warnings
    # import warnings
    # warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='py_exp')
    res_arr = numeric.quantilize_without_nan(df.values, n_quantiles=n_quantiles, axis=axis)
    if isinstance(df, pd.DataFrame):
        res = pd.DataFrame(index=df.index, columns=df.columns, data=res_arr)
    elif isinstance(df, pd.Series):
        res = pd.Series(index=df.index, data=res_arr)
    else:
        raise ValueError
    return res


def fillinf(df):
    return df.replace([np.inf, -np.inf], np.nan)


def group_df_to_dict(df, by):
    gp = df.groupby(by=by)
    res = {key: value for key, value in gp}
    return res


def rank_with_mask(df, axis=1, mask=None, normalize=False, method='min'):
    """
    
    Parameters
    ----------
    df : pd.DataFrame
    axis : {0, 1}
    mask : pd.DataFrame
    normalize : bool
    method : {'min', 'average', 'max', 'dense'}

    Returns
    -------
    pd.DataFrame
    
    Notes
    -----
    If calculate rank, use 'min' method by default;
    If normalize, result will range in [0.0, 1.0]

    """
    not_nan_mask = (~df.isnull())
    
    if mask is None:
        mask = not_nan_mask
    else:
        mask = np.logical_and(not_nan_mask, mask)
    
    rank = df[mask].rank(axis=axis, na_option='keep', method=method)
    
    if normalize:
        dividend = rank.max(axis=axis)
        SUB = 1
        # for dividend = 1, do not subtract 1, otherwise there will be NaN
        dividend.loc[dividend > SUB] = dividend.loc[dividend > SUB] - SUB
        rank = rank.sub(SUB).div(dividend, axis=(1 - axis))
    return rank



# encoding: utf-8
"""


"""
from __future__ import print_function
import os
try:
    basestring
except NameError:
    basestring = str

import numpy as np
import pandas as pd

import jaqs.util as jutil
from jaqs.data.align import align
from jaqs.data.py_expression_eval import Parser




class DataView(object):
    """
    Prepare data before research / trade. Support file I/O.
    Support: add field, add formula, save / load.
    
    Attributes
    ----------
    data_api : RemoteDataService
    symbol : list
    start_date : int
    end_date : int
    fields : list
    freq : int
    market_daily_fields, reference_daily_fields : list
    data_d : pd.DataFrame
        All daily frequency data will be merged and stored here.
        index is date, columns is symbol-field MultiIndex
    data_q : pd.DataFrame
        All quarterly frequency data will be merged and stored here.
        index is date, columns is symbol-field MultiIndex
    
    """
    def __init__(self):
        self.data_api = None
        
        self.universe = ""
        self.symbol = []
        self.benchmark = ""
        self.start_date = 0
        self.extended_start_date_d = 0
        self.extended_start_date_q = 0
        self.end_date = 0
        self.fields = []
        self.freq = 1
        self.all_price = True

        self.meta_data_list = ['start_date', 'end_date',
                               'extended_start_date_d', 'extended_start_date_q',
                               'freq', 'fields', 'symbol', 'universe', 'benchmark',
                               'custom_daily_fields', 'custom_quarterly_fields']
        self.adjust_mode = 'post'
        
        self.data_d = None
        self.data_q = None
        self._data_benchmark = None
        self._data_inst = None
        # self._data_group = None
        
        common_list = {'symbol', 'start_date', 'end_date'}
        market_bar_list = {'open', 'high', 'low', 'close', 'volume', 'turnover', 'vwap', 'oi'}
        market_tick_list = {'volume', 'oi',
                            'askprice1', 'askprice2', 'askprice3', 'askprice4', 'askprice5',
                            'bidprice1', 'bidprice1', 'bidprice1', 'bidprice1', 'bidprice1',
                            'askvolume1', 'askvolume2', 'askvolume3', 'askvolume4', 'askvolume5',
                            'bidvolume1', 'bidvolume2', 'bidvolume3', 'bidvolume4', 'bidvolume5'}
        # fields map
        # TODO: 'freq' is not in market_daily_fields yet.
        self.market_daily_fields = \
            {'open', 'high', 'low', 'close', 'volume', 'turnover', 'vwap', 'oi', 'trade_status',
             'open_adj', 'high_adj', 'low_adj', 'close_adj', 'vwap_adj', 'index_member', 'index_weight'}
        self.group_fields = {'sw1', 'sw2', 'sw3', 'sw4', 'zz1', 'zz2'}
        self.reference_daily_fields = \
            {"total_mv", "float_mv", "pe", "pb", "pe_ttm", "pcf_ocf", "pcf_ocfttm", "pcf_ncf",
             "pcf_ncfttm", "ps", "ps_ttm", "turnover_ratio", "free_turnover_ratio", "total_share",
             "float_share", "price_div_dps", "free_share", "np_parent_comp_ttm",
             "np_parent_comp_lyr", "net_assets", "ncf_oper_ttm", "ncf_oper_lyr", "oper_rev_ttm",
             "oper_rev_lyr", "limit_status"}
        self.fin_stat_income = \
            {"symbol", "ann_date", "start_date", "end_date",
             "comp_type_code", "comp_type_code", "act_ann_date", "start_actdate",
             "end_actdate", "report_date", "start_reportdate", "start_reportdate",
             "report_type", "report_type", "currency", "total_oper_rev", "oper_rev",
             "int_income", "net_int_income", "insur_prem_unearned",
             "handling_chrg_income", "net_handling_chrg_income", "net_inc_other_ops",
             "plus_net_inc_other_bus", "prem_income", "less_ceded_out_prem",
             "chg_unearned_prem_res", "incl_reinsurance_prem_inc",
             "net_inc_sec_trading_brok_bus", "net_inc_sec_uw_bus",
             "net_inc_ec_asset_mgmt_bus", "other_bus_income", "plus_net_gain_chg_fv",
             "plus_net_invest_inc", "incl_inc_invest_assoc_jv_entp",
             "plus_net_gain_fx_trans", "tot_oper_cost", "less_oper_cost", "less_int_exp",
             "less_handling_chrg_comm_exp", "less_taxes_surcharges_ops",
             "less_selling_dist_exp", "less_gerl_admin_exp", "less_fin_exp",
             "less_impair_loss_assets", "prepay_surr", "tot_claim_exp",
             "chg_insur_cont_rsrv", "dvd_exp_insured", "reinsurance_exp", "oper_exp",
             "less_claim_recb_reinsurer", "less_ins_rsrv_recb_reinsurer",
             "less_exp_recb_reinsurer", "other_bus_cost", "oper_profit",
             "plus_non_oper_rev", "less_non_oper_exp", "il_net_loss_disp_noncur_asset",
             "tot_profit", "inc_tax", "unconfirmed_invest_loss",
             "net_profit_incl_min_int_inc", "net_profit_excl_min_int_inc",
             "minority_int_inc", "other_compreh_inc", "tot_compreh_inc",
             "tot_compreh_inc_parent_comp", "tot_compreh_inc_min_shrhldr", "ebit",
             "ebitda", "net_profit_after_ded_nr_lp", "net_profit_under_intl_acc_sta",
             "eps_basic", "eps_diluted", "insurance_expense",
             "spe_bal_oper_profit", "tot_bal_oper_profit", "spe_bal_tot_profit",
             "tot_bal_tot_profit", "spe_bal_net_profit", "tot_bal_net_profit",
             "undistributed_profit", "adjlossgain_prevyear",
             "transfer_from_surplusreserve", "transfer_from_housingimprest",
             "transfer_from_others", "distributable_profit", "withdr_legalsurplus",
             "withdr_legalpubwelfunds", "workers_welfare", "withdr_buzexpwelfare",
             "withdr_reservefund", "distributable_profit_shrhder", "prfshare_dvd_payable",
             "withdr_othersurpreserve", "comshare_dvd_payable",
             "capitalized_comstock_div"}
        self.fin_stat_balance_sheet = \
            {"monetary_cap", "tradable_assets", "notes_rcv", "acct_rcv", "other_rcv", "pre_pay",
             "dvd_rcv", "int_rcv", "inventories", "consumptive_assets", "deferred_exp",
             "noncur_assets_due_1y", "settle_rsrv", "loans_to_banks", "prem_rcv", "rcv_from_reinsurer",
             "rcv_from_ceded_insur_cont_rsrv", "red_monetary_cap_for_sale", "other_cur_assets",
             "tot_cur_assets", "fin_assets_avail_for_sale", "held_to_mty_invest", "long_term_eqy_invest",
             "invest_real_estate", "time_deposits", "other_assets", "long_term_rec", "fix_assets",
             "const_in_prog", "proj_matl", "fix_assets_disp", "productive_bio_assets",
             "oil_and_natural_gas_assets", "intang_assets", "r_and_d_costs", "goodwill",
             "long_term_deferred_exp", "deferred_tax_assets", "loans_and_adv_granted",
             "oth_non_cur_assets", "tot_non_cur_assets", "cash_deposits_central_bank",
             "asset_dep_oth_banks_fin_inst", "precious_metals", "derivative_fin_assets",
             "agency_bus_assets", "subr_rec", "rcv_ceded_unearned_prem_rsrv", "rcv_ceded_claim_rsrv",
             "rcv_ceded_life_insur_rsrv", "rcv_ceded_lt_health_insur_rsrv", "mrgn_paid",
             "insured_pledge_loan", "cap_mrgn_paid", "independent_acct_assets", "clients_cap_deposit",
             "clients_rsrv_settle", "incl_seat_fees_exchange", "rcv_invest", "tot_assets", "st_borrow",
             "borrow_central_bank", "deposit_received_ib_deposits", "loans_oth_banks", "tradable_fin_liab",
             "notes_payable", "acct_payable", "adv_from_cust", "fund_sales_fin_assets_rp",
             "handling_charges_comm_payable", "empl_ben_payable", "taxes_surcharges_payable", "int_payable",
             "dvd_payable", "other_payable", "acc_exp", "deferred_inc", "st_bonds_payable", "payable_to_reinsurer",
             "rsrv_insur_cont", "acting_trading_sec", "acting_uw_sec", "non_cur_liab_due_within_1y", "other_cur_liab",
             "tot_cur_liab", "lt_borrow", "bonds_payable", "lt_payable", "specific_item_payable", "provisions",
             "deferred_tax_liab", "deferred_inc_non_cur_liab", "other_non_cur_liab", "tot_non_cur_liab",
             "liab_dep_other_banks_inst", "derivative_fin_liab", "cust_bank_dep", "agency_bus_liab", "other_liab",
             "prem_received_adv", "deposit_received", "insured_deposit_invest", "unearned_prem_rsrv", "out_loss_rsrv",
             "life_insur_rsrv", "lt_health_insur_v", "independent_acct_liab", "incl_pledge_loan", "claims_payable",
             "dvd_payable_insured", "total_liab", "capital_stk", "capital_reser", "special_rsrv", "surplus_rsrv",
             "undistributed_profit", "less_tsy_stk", "prov_nom_risks", "cnvd_diff_foreign_curr_stat",
             "unconfirmed_invest_loss", "minority_int", "tot_shrhldr_eqy_excl_min_int", "tot_shrhldr_eqy_incl_min_int",
             "tot_liab_shrhldr_eqy", "spe_cur_assets_diff", "tot_cur_assets_diff", "spe_non_cur_assets_diff",
             "tot_non_cur_assets_diff", "spe_bal_assets_diff", "tot_bal_assets_diff", "spe_cur_liab_diff",
             "tot_cur_liab_diff", "spe_non_cur_liab_diff", "tot_non_cur_liab_diff", "spe_bal_liab_diff",
             "tot_bal_liab_diff", "spe_bal_shrhldr_eqy_diff", "tot_bal_shrhldr_eqy_diff", "spe_bal_liab_eqy_diff",
             "tot_bal_liab_eqy_diff", "lt_payroll_payable", "other_comp_income", "other_equity_tools",
             "other_equity_tools_p_shr", "lending_funds", "accounts_receivable", "st_financing_payable", "payables"}
        self.fin_stat_cash_flow = \
            {"cash_recp_sg_and_rs", "recp_tax_rends", "net_incr_dep_cob", "net_incr_loans_central_bank",
             "net_incr_fund_borr_ofi", "cash_recp_prem_orig_inco", "net_incr_insured_dep",
             "net_cash_received_reinsu_bus", "net_incr_disp_tfa", "net_incr_int_handling_chrg", "net_incr_disp_faas",
             "net_incr_loans_other_bank", "net_incr_repurch_bus_fund", "other_cash_recp_ral_oper_act",
             "stot_cash_inflows_oper_act", "cash_pay_goods_purch_serv_rec", "cash_pay_beh_empl", "pay_all_typ_tax",
             "net_incr_clients_loan_adv", "net_incr_dep_cbob", "cash_pay_claims_orig_inco", "handling_chrg_paid",
             "comm_insur_plcy_paid", "other_cash_pay_ral_oper_act", "stot_cash_outflows_oper_act",
             "net_cash_flows_oper_act", "cash_recp_disp_withdrwl_invest", "cash_recp_return_invest",
             "net_cash_recp_disp_fiolta", "net_cash_recp_disp_sobu", "other_cash_recp_ral_inv_act",
             "stot_cash_inflows_inv_act", "cash_pay_acq_const_fiolta", "cash_paid_invest", "net_cash_pay_aquis_sobu",
             "other_cash_pay_ral_inv_act", "net_incr_pledge_loan", "stot_cash_outflows_inv_act",
             "net_cash_flows_inv_act", "cash_recp_cap_contrib", "incl_cash_rec_saims", "cash_recp_borrow",
             "proc_issue_bonds", "other_cash_recp_ral_fnc_act", "stot_cash_inflows_fnc_act", "cash_prepay_amt_borr",
             "cash_pay_dist_dpcp_int_exp", "incl_dvd_profit_paid_sc_ms", "other_cash_pay_ral_fnc_act",
             "stot_cash_outflows_fnc_act", "net_cash_flows_fnc_act", "eff_fx_flu_cash", "net_incr_cash_cash_equ",
             "cash_cash_equ_beg_period", "cash_cash_equ_end_period", "net_profit", "unconfirmed_invest_loss",
             "plus_prov_depr_assets", "depr_fa_coga_dpba", "amort_intang_assets", "amort_lt_deferred_exp",
             "decr_deferred_exp", "incr_acc_exp", "loss_disp_fiolta", "loss_scr_fa", "loss_fv_chg", "fin_exp",
             "invest_loss", "decr_deferred_inc_tax_assets", "incr_deferred_inc_tax_liab", "decr_inventories",
             "decr_oper_payable", "incr_oper_payable", "others", "im_net_cash_flows_oper_act", "conv_debt_into_cap",
             "conv_corp_bonds_due_within_1y", "fa_fnc_leases", "end_bal_cash", "less_beg_bal_cash",
             "plus_end_bal_cash_equ", "less_beg_bal_cash_equ", "im_net_incr_cash_cash_equ", "free_cash_flow",
             "spe_bal_cash_inflows_oper", "tot_bal_cash_inflows_oper", "spe_bal_cash_outflows_oper",
             "tot_bal_cash_outflows_oper", "tot_bal_netcash_outflows_oper", "spe_bal_cash_inflows_inv",
             "tot_bal_cash_inflows_inv", "spe_bal_cash_outflows_inv", "tot_bal_cash_outflows_inv",
             "tot_bal_netcash_outflows_inv", "spe_bal_cash_inflows_fnc", "tot_bal_cash_inflows_fnc",
             "spe_bal_cash_outflows_fnc", "tot_bal_cash_outflows_fnc", "tot_bal_netcash_outflows_fnc",
             "spe_bal_netcash_inc", "tot_bal_netcash_inc", "spe_bal_netcash_equ_undir", "tot_bal_netcash_equ_undir",
             "spe_bal_netcash_inc_undir", "tot_bal_netcash_inc_undir"}
        self.fin_indicator = \
            {"extraordinary","deductedprofit","grossmargin","operateincome","investincome","stmnote_finexp",
             "stm_is","ebit_daily","ebitda""fcff","fcfe","exinterestdebt_current","exinterestdebt_noncurrent","interestdebt",
             "netdebt","tangibleasset","workingcapital","networkingcapital","investcapital","retainedearnings","eps_basic_daily", # TODO eps_basic
             "eps_diluted","eps_diluted2","bps","ocfps","grps","orps","surpluscapitalps","surplusreserveps","undistributedps",
             "retainedps","cfps","ebitps","fcffps","fcfeps","netprofitmargin","grossprofitmargin","cogstosales",
             "expensetosales","profittogr","saleexpensetogr","adminexpensetogr","finaexpensetogr","impairtogr_ttm",
             "gctogr","optogr","ebittogr","roe","roe_deducted","roa2","roa","roic","roe_yearly","roa2_yearly","roe_avg",
             "operateincometoebt","investincometoebt","nonoperateprofittoebt","taxtoebt","deductedprofittoprofit","salescashintoor",
             "ocftoor","ocftooperateincome","capitalizedtoda","debttoassets","assetstoequity","dupont_assetstoequity",
             "catoassets","ncatoassets","tangibleassetstoassets","intdebttototalcap","equitytototalcapital","currentdebttodebt",
             "longdebtodebt","current","quick","cashratio","ocftoshortdebt","debttoequity","equitytodebt",
             "equitytointerestdebt","tangibleassettodebt","tangassettointdebt","tangibleassettonetdebt","ocftodebt",
             "ocftointerestdebt","ocftonetdebt","ebittointerest","longdebttoworkingcapital","ebitdatodebt","turndays",
             "invturndays","arturndays","invturn","arturn","caturn","faturn","assetsturn","roa_yearly","dupont_roa",
             "s_stm_bs","prefinexpense_opprofit","nonopprofit","optoebt","noptoebt","ocftoprofit","cashtoliqdebt",
             "cashtoliqdebtwithinterest","optoliqdebt","optodebt","roic_yearly","tot_faturn","profittoop","qfa_operateincome",
             "qfa_investincome","qfa_deductedprofit","qfa_eps","qfa_netprofitmargin","qfa_grossprofitmargin","qfa_expensetosales",
             "qfa_profittogr","qfa_saleexpensetogr","qfa_adminexpensetogr","qfa_finaexpensetogr","qfa_impairtogr_ttm",
             "qfa_gctogr","qfa_optogr","qfa_roe","qfa_roe_deducted","qfa_roa","qfa_operateincometoebt","qfa_investincometoebt",
             "qfa_deductedprofittoprofit","qfa_salescashintoor","qfa_ocftosales","qfa_ocftoor","yoyeps_basic","yoyeps_diluted",
             "yoyocfps","yoyop","yoyebt","yoynetprofit","yoynetprofit_deducted","yoyocf","yoyroe","yoybps","yoyassets",
             "yoyequity","yoy_tr","yoy_or","qfa_yoygr","qfa_cgrgr","qfa_yoysales","qfa_cgrsales","qfa_yoyop","qfa_cgrop",
             "qfa_yoyprofit","qfa_cgrprofit","qfa_yoynetprofit","qfa_cgrnetprofit","yoy_equity","rd_expense","waa_roe"}
        self .custom_daily_fields = []
        self .custom_quarterly_fields = []
        
        # co nst
        self .ANN_DATE_FIELD_NAME = 'ann_date'
        self .REPORT_DATE_FIELD_NAME = 'report_date'
        self.TRADE_STATUS_FIELD_NAME = 'trade_status'
        self.TRADE_DATE_FIELD_NAME = 'trade_date'
    
    # --------------------------------------------------------------------------------------------------------
    # Properties
    @property
    def data_benchmark(self):
        return self._data_benchmark
    
    @property
    def data_inst(self):
        """
        
        Returns
        -------
        pd.DataFrame

        """
        return self._data_inst
    
    @data_benchmark.setter
    def data_benchmark(self, df_new):
        if self.data_d is not None and df_new.shape[0] != self.data_d.shape[0]:
            raise ValueError("You must provide a DataFrame with the same shape of data_benchmark.")
        self._data_benchmark = df_new

    @property
    def dates(self):
        """
        Get daily date array of the underlying data.
        
        Returns
        -------
        res : np.array
            dtype: int

        """
        if self.data_d is not None:
            res = self.data_d.index.values
        elif self.data_api is not None:
            res = self.data_api.query_trade_dates(self.extended_start_date_d, self.end_date)
        else:
            raise ValueError("Cannot get dates array when neither of data and data_api exists.")
    
        return res

    # --------------------------------------------------------------------------------------------------------
    # Fields
    def _is_quarter_field(self, field_name):
        """
        Check whether a field name is quarterly frequency.
        field_name must be pre-defined or already added.
        
        Parameters
        ----------
        field_name : str

        Returns
        -------
        bool

        """
        res = (field_name in self.fin_stat_balance_sheet
               or field_name in self.fin_stat_cash_flow
               or field_name in self.fin_stat_income
               or field_name in self.fin_indicator
               or field_name in self.custom_quarterly_fields)
        return res

    def _is_daily_field(self, field_name):
        """
        Check whether a field name is daily frequency.
        field_name must be pre-defined or already added.
        
        Parameters
        ----------
        field_name : str

        Returns
        -------
        bool

        """
        flag = (field_name in self.market_daily_fields
                or field_name in self.reference_daily_fields
                or field_name in self.custom_daily_fields
                or field_name in self.group_fields)
        return flag

    def _is_predefined_field(self, field_name):
        """
        Check whether a field name can be recognized.
        field_name must be pre-defined or already added.
        
        Parameters
        ----------
        field_name : str

        Returns
        -------
        bool

        """
        return self._is_quarter_field(field_name) or self._is_daily_field(field_name)

    def _get_fields(self, field_type, fields, complement=False, append=False):
        """
        Get list of fields that are in ref_quarterly_fields.
        
        Parameters
        ----------
        field_type : {'market_daily', 'ref_daily', 'income', 'balance_sheet', 'cash_flow', 'daily', 'quarterly'
        fields : list of str
        complement : bool, optional
            If True, get fields that are NOT in ref_quarterly_fields.

        Returns
        -------
        list

        """
        pool_map = {'market_daily': self.market_daily_fields,
                    'ref_daily': self.reference_daily_fields,
                    'income': self.fin_stat_income,
                    'balance_sheet': self.fin_stat_balance_sheet,
                    'cash_flow': self.fin_stat_cash_flow,
                    'fin_indicator': self.fin_indicator,
                    'group': self.group_fields}
        pool_map['daily'] = set.union(pool_map['market_daily'],
                                      pool_map['ref_daily'],
                                      pool_map['group'],
                                      self.custom_daily_fields)
        pool_map['quarterly'] = set.union(pool_map['income'],
                                          pool_map['balance_sheet'],
                                          pool_map['cash_flow'],
                                          pool_map['fin_indicator'],
                                          self.custom_quarterly_fields)
        
        pool = pool_map.get(field_type, None)
        if pool is None:
            raise NotImplementedError("field_type = {:s}".format(field_type))
        
        s = set.intersection(set(pool), set(fields))
        if not s:
            return []
        
        if complement:
            s = set(fields) - s
            
        if field_type == 'market_daily' and self.all_price:
            # turnover will not be adjusted
            s.update({'open', 'high', 'close', 'low', 'vwap'})
            
        if append:
            s.add('symbol')
            if field_type == 'market_daily' or field_type == 'ref_daily':
                s.add('trade_date')
                if field_type == 'market_daily':
                    s.add(self.TRADE_STATUS_FIELD_NAME)
            elif (field_type == 'income'
                  or field_type == 'balance_sheet'
                  or field_type == 'cash_flow'
                  or field_type == 'fin_indicator'):
                s.add(self.ANN_DATE_FIELD_NAME)
                s.add(self.REPORT_DATE_FIELD_NAME)
        
        l = list(s)
        return l

    # --------------------------------------------------------------------------------------------------------
    # Prepare data
    def init_from_config(self, props, data_api):
        """
        Initialize various attributes like start/end date, universe/symbol, fields, etc.
        If your want to parse symbol, but use a custom benchmark index, please directly assign self.data_benchmark.
        
        Parameters
        ----------
        props : dict
            start_date, end_date, freq, symbol, fields, etc.
        data_api : BaseDataServer
        
        """
        # data_api.init_from_config(props)
        self.data_api = data_api
    
        sep = ','
    
        # initialize parameters
        self.start_date = props['start_date']
        self.extended_start_date_d = jutil.shift(self.start_date, n_weeks=-8)  # query more data
        self.extended_start_date_q = jutil.shift(self.start_date, n_weeks=-80)
        self.end_date = props['end_date']
        self.all_price = props.get('all_price', True)
        self.freq = props.get('freq', 1)
    
        # get and filter fields
        fields = props.get('fields', [])
        if fields:
            fields = props['fields'].split(sep)
            self.fields = [field for field in fields if self._is_predefined_field(field)]
            if len(self.fields) < len(fields):
                print("Field name [{}] not valid, ignore.".format(set.difference(set(fields), set(self.fields))))
    
        # append additional fields
        if self.all_price:
            self.fields.extend(['open_adj', 'high_adj', 'low_adj', 'close_adj',
                                'open', 'high', 'low', 'close',
                                'vwap', 'vwap_adj'])
    
        # initialize universe/symbol
        universe = props.get('universe', "")
        symbol = props.get('symbol', "")
        benchmark = props.get('benchmark', '')
        if symbol and universe:
            raise ValueError("Please use either [symbol] or [universe].")
        if not (symbol or universe):
            raise ValueError("One of [symbol] or [universe] must be provided.")
        if universe:
            univ_list = universe.split(',')
            self.universe = univ_list
            symbols_list = []
            for univ in univ_list:
                symbols_list.extend(data_api.query_index_member(univ, self.extended_start_date_d, self.end_date))
            self.symbol = sorted(list(set(symbols_list)))
        else:
            self.symbol = sorted(symbol.split(sep))
        if benchmark:
            self.benchmark = benchmark
        else:
            if self.universe:
                if len(self.universe) > 1:
                    print("More than one universe are used: {}, "
                                     "use the first one ({}) as index by default. "
                                     "If you want to use other benchmark, "
                                     "please specify benchmark in configs.".format(repr(self.universe), self.universe[0]))
                self.benchmark = self.universe[0]
    
        print("Initialize config success.")

    def distributed_query(self, query_func_name, symbol, start_date, end_date, limit=100000, **kwargs):
        n_symbols = len(symbol.split(','))
        dates = self.data_api.query_trade_dates(start_date, end_date)
        n_days = len(dates)
        
        if n_symbols * n_days > limit:
            n = limit // n_symbols
            
            df_list = []
            i = 0
            pos1, pos2 = n * i, n * (i + 1) - 1
            while pos2 < n_days:
                print(pos2)
                df, msg = getattr(self.data_api, query_func_name)(symbol=symbol,
                                                                  start_date=dates[pos1], end_date=dates[pos2],
                                                                  **kwargs)
                df_list.append(df)
                i += 1
                pos1, pos2 = n * i, n * (i + 1) - 1
            if pos1 < n_days:
                df, msg = getattr(self.data_api, query_func_name)(symbol=symbol,
                                                                  start_date=dates[pos1], end_date=dates[-1],
                                                                  **kwargs)
                df_list.append(df)
            df = pd.concat(df_list, axis=0)
        else:
            df, msg = getattr(self.data_api, query_func_name)(symbol,
                                                              start_date=start_date, end_date=end_date,
                                                              **kwargs)
        return df, msg

    def prepare_data(self):
        """Prepare data for the FIRST time."""
        # prepare benchmark and group
        print("Query data...")
        data_d, data_q = self._prepare_daily_quarterly(self.fields)
        self.data_d, self.data_q = data_d, data_q
        if self.data_q is not None:
            self._prepare_report_date()
        self._align_and_merge_q_into_d()
        
        print("Query instrument info...")
        self._prepare_inst_info()
    
        print("Query adj_factor...")
        self._prepare_adj_factor()
    
        if self.benchmark:
            print("Query benchmark...")
            self._data_benchmark = self._prepare_benchmark()
        if self.universe:
            print("Query benchmar member info...")
            self._prepare_comp_info()
    
        group_fields = self._get_fields('group', self.fields)
        if group_fields:
            print("Query groups (industry)...")
            self._prepare_group(group_fields)
    
        print("Data has been successfully prepared.")

    @staticmethod
    def _process_index_co(df, index_name):
        df = df.astype(dtype={index_name: int})
        df = df.drop_duplicates(subset=['symbol', index_name])
        return df

    def _prepare_daily_quarterly(self, fields):
        """
        Query and process data from data_api.
        
        Parameters
        ----------
        fields : list

        Returns
        -------
        merge_d : pd.DataFrame or None
        merge_q : pd.DataFrame or None

        """
        if not fields:
            return None, None
    
        # query data
        print("Query data - query...")
        daily_list, quarterly_list = self._query_data(self.symbol, fields)
    
        def pivot_and_sort(df, index_name):
            df = self._process_index_co(df, index_name)
            df = df.pivot(index=index_name, columns='symbol')
            df.columns = df.columns.swaplevel()
            col_names = ['symbol', 'field']
            df.columns.names = col_names
            df = df.sort_index(axis=1, level=col_names)
            df.index.name = index_name
            return df
    
        multi_daily = None
        multi_quarterly = None
        if daily_list:
            daily_list_pivot = [pivot_and_sort(df, self.TRADE_DATE_FIELD_NAME) for df in daily_list]
            multi_daily = self._merge_data(daily_list_pivot, self.TRADE_DATE_FIELD_NAME)
            # use self.dates as index because original data have weekends
            multi_daily = self._fill_missing_idx_col(multi_daily, index=self.dates, symbols=self.symbol)
            print("Query data - daily fields prepared.")
        if quarterly_list:
            quarterly_list_pivot = [pivot_and_sort(df, self.REPORT_DATE_FIELD_NAME) for df in quarterly_list]
            multi_quarterly = self._merge_data(quarterly_list_pivot, self.REPORT_DATE_FIELD_NAME)
            multi_quarterly = self._fill_missing_idx_col(multi_quarterly, index=None, symbols=self.symbol)
            print("Query data - quarterly fields prepared.")
    
        return multi_daily, multi_quarterly

    def _query_data(self, symbol, fields):
        """
        Query data using different APIs, then store them in dict.
        period, start_date and end_date are fixed.
        Keys of dict are securitites.
        
        Parameters
        ----------
        symbol : list of str
        fields : list of str

        Returns
        -------
        daily_list : list
        quarterly_list : list

        """
        sep = ','
        symbol_str = sep.join(symbol)
    
        if self.freq == 1:
            daily_list = []
            quarterly_list = []
        
            # TODO : use fields = {field: kwargs} to enable params
            fields_market_daily = self._get_fields('market_daily', fields, append=True)
            if fields_market_daily:
                print("NOTE: price adjust method is [{:s} adjust]".format(self.adjust_mode))
                # no adjust prices and other market daily fields
                df_daily, msg1 = self.distributed_query('daily', symbol_str,
                                                        start_date=self.extended_start_date_d, end_date=self.end_date,
                                                        adjust_mode=None, fields=sep.join(fields_market_daily), limit=100000)
                #df_daily, msg1 = self.data_api.daily(symbol_str, start_date=self.extended_start_date_d, end_date=self.end_date,
                #                                     adjust_mode=None, fields=sep.join(fields_market_daily))
            
                if self.all_price:
                    adj_cols = ['open', 'high', 'low', 'close', 'vwap']
                    # adjusted prices
                    #df_daily_adjust, msg11 = self.data_api.daily(symbol_str, start_date=self.extended_start_date_d, end_date=self.end_date,
                    #                                             adjust_mode=self.adjust_mode, fields=','.join(adj_cols))
                    df_daily_adjust, msg1 = self.distributed_query('daily', symbol_str,
                                                                   start_date=self.extended_start_date_d, end_date=self.end_date,
                                                                   adjust_mode=self.adjust_mode, fields=sep.join(fields_market_daily), limit=100000)
                
                    df_daily = pd.merge(df_daily, df_daily_adjust, how='outer',
                                        on=['symbol', 'trade_date'], suffixes=('', '_adj'))
                daily_list.append(df_daily.loc[:, fields_market_daily])
        
            fields_ref_daily = self._get_fields('ref_daily', fields, append=True)
            if fields_ref_daily:
                df_ref_daily, msg2 = self.distributed_query('query_lb_dailyindicator', symbol_str,
                                                            start_date=self.extended_start_date_d, end_date=self.end_date,
                                                            fields=sep.join(fields_ref_daily), limit=20000)
                daily_list.append(df_ref_daily.loc[:, fields_ref_daily])
        
            fields_income = self._get_fields('income', fields, append=True)
            if fields_income:
                df_income, msg3 = self.data_api.query_lb_fin_stat('income', symbol_str, self.extended_start_date_q, self.end_date,
                                                                  sep.join(fields_income), drop_dup_cols=['symbol', self.REPORT_DATE_FIELD_NAME])
                quarterly_list.append(df_income.loc[:, fields_income])
        
            fields_balance = self._get_fields('balance_sheet', fields, append=True)
            if fields_balance:
                df_balance, msg3 = self.data_api.query_lb_fin_stat('balance_sheet', symbol_str, self.extended_start_date_q, self.end_date,
                                                                   sep.join(fields_balance), drop_dup_cols=['symbol', self.REPORT_DATE_FIELD_NAME])
                quarterly_list.append(df_balance.loc[:, fields_balance])
        
            fields_cf = self._get_fields('cash_flow', fields, append=True)
            if fields_cf:
                df_cf, msg3 = self.data_api.query_lb_fin_stat('cash_flow', symbol_str, self.extended_start_date_q, self.end_date,
                                                              sep.join(fields_cf), drop_dup_cols=['symbol', self.REPORT_DATE_FIELD_NAME])
                quarterly_list.append(df_cf.loc[:, fields_cf])
        
            fields_fin_ind = self._get_fields('fin_indicator', fields, append=True)
            if fields_fin_ind:
                df_fin_ind, msg4 = self.data_api.query_lb_fin_stat('fin_indicator', symbol_str,
                                                                   self.extended_start_date_q, self.end_date,
                                                                   sep.join(fields_fin_ind), drop_dup_cols=['symbol', self.REPORT_DATE_FIELD_NAME])
                quarterly_list.append(df_fin_ind.loc[:, fields_fin_ind])
    
        else:
            raise NotImplementedError("freq = {}".format(self.freq))
    
        return daily_list, quarterly_list

    @staticmethod
    def _merge_data(dfs, index_name='trade_date'):
        """
        Merge data from different APIs into one DataFrame.
        
        Parameters
        ----------
        dfs : list of pd.DataFrame

        Returns
        -------
        merge : pd.DataFrame or None
            If dfs is empty, return None
        
        Notes
        -----
        Align on date index, concatenate on columns (symbol and fields)
        
        """
        # dfs = [df for df in dfs if df is not None]
    
        merge = pd.concat(dfs, axis=1, join='outer')
    
        # drop duplicated columns. ONE LINE EFFICIENT version
        mask_duplicated = merge.columns.duplicated()
        if np.any(mask_duplicated):
            # print("Duplicated columns found. Dropped.")
            merge = merge.loc[:, ~mask_duplicated]
        
            # if merge.isnull().sum().sum() > 0:
            # print "WARNING: nan in final merged data. NO fill"
            # merge.fillna(method='ffill', inplace=True)
    
        merge = merge.sort_index(axis=1, level=['symbol', 'field'])
        merge.index.name = index_name
    
        return merge

    def _fill_missing_idx_col(self, df, index=None, symbols=None):
        if index is None:
            index = df.index
        if symbols is None:
            symbols = self.symbol
        fields = df.columns.levels[1]
    
        if len(fields) * len(self.symbol) != len(df.columns) or len(index) != len(df.index):
            cols_multi = pd.MultiIndex.from_product([symbols, fields], names=['symbol', 'field'])
            cols_multi = cols_multi.sort_values()
            df_final = pd.DataFrame(index=index, columns=cols_multi, data=np.nan)
            df_final.index.name = df.index.name
        
            df_final.update(df)
        
            # idx_diff = sorted(set(df_final.index) - set(df.index))
            col_diff = sorted(set(df_final.columns.levels[0].values) - set(df.columns.levels[0].values))
            print ("WARNING: some data is unavailable: "
                   # + "\n    At index " + ', '.join(idx_diff)
                   + "\n    At fields " + ', '.join(col_diff))
            return df_final
        else:
            return df

    def _align_and_merge_q_into_d(self):
        data_d, data_q = self.data_d, self.data_q
        if data_d is not None and data_q is not None:
            df_ref_ann = data_q.loc[:, pd.IndexSlice[:, self.ANN_DATE_FIELD_NAME]].copy()
            df_ref_ann.columns = df_ref_ann.columns.droplevel(level='field')
        
            dic_expanded = dict()
            for field_name, df in data_q.groupby(level=1, axis=1):  # by column multiindex fields
                df_expanded = align(df, df_ref_ann, self.dates)
                dic_expanded[field_name] = df_expanded
            df_quarterly_expanded = pd.concat(dic_expanded.values(), axis=1)
            df_quarterly_expanded.index.name = self.TRADE_DATE_FIELD_NAME
        
            data_d_merge = self._merge_data([data_d, df_quarterly_expanded], index_name=self.TRADE_DATE_FIELD_NAME)
            data_d = data_d_merge.loc[data_d.index, :]
        self.data_d = data_d

    def _prepare_adj_factor(self):
        """Query and append daily adjust factor for prices."""
        mask_stocks = self.data_inst['inst_type'] == 1
        if mask_stocks.sum() == 0:
            return
        symbol_stocks = self.data_inst.loc[mask_stocks].index.values
        symbol_str = ','.join(symbol_stocks)
        df_adj = self.data_api.query_adj_factor_daily(symbol_str,
                                                      start_date=self.extended_start_date_d, end_date=self.end_date, div=False)
        self.append_df(df_adj, 'adjust_factor', is_quarterly=False)

    def _prepare_comp_info(self):
        # if a symbol is index member of any one universe, its value of index_member will be 1.0
        res = dict()
        for univ in self.universe:
            df = self.data_api.query_index_member_daily(univ, self.extended_start_date_d, self.end_date)
            res[univ] = df
        df_res = pd.concat(res, axis=0)
        df = df_res.groupby(by='trade_date').apply(lambda df: df.any(axis=0)).astype(float)
        self.append_df(df, 'index_member', is_quarterly=False)
    
        # use weights of the first universe
        df_weights = self.data_api.query_index_weights_daily(self.universe[0], self.extended_start_date_d, self.end_date)
        self.append_df(df_weights, 'index_weight', is_quarterly=False)

    def _prepare_report_date(self):
        idx = self.data_q.index
        df_report_date = pd.DataFrame(index=idx, columns=self.symbol, data=0)
        n = len(idx)
        quarter = idx.values // 100 % 100
        df_report_date.loc[:, :] = quarter.reshape(n, -1)
        
        self.append_df(df_report_date, 'quarter', is_quarterly=True)
    
    def _prepare_inst_info(self):
        res = self.data_api.query_inst_info(symbol=','.join(self.symbol),
                                            fields='symbol,inst_type,name,list_date,'
                                                   'delist_date,product,pricetick,multiplier,'
                                                   'buylot,setlot',
                                            inst_type="")
        self._data_inst = res

    def _prepare_group(self, group_fields):
        data_map = {'sw1': ('SW', 1),
                    'sw2': ('SW', 2),
                    'sw3': ('SW', 3),
                    'sw4': ('SW', 4),
                    'zz1': ('ZZ', 1),
                    'zz2': ('ZZ', 2)}
        for field in group_fields:
            type_, level = data_map[field]
            df = self.data_api.query_industry_daily(symbol=','.join(self.symbol),
                                                    start_date=self.extended_start_date_q, end_date=self.end_date,
                                                    type_=type_, level=level)
            self.append_df(df, field, is_quarterly=False)

    def _prepare_benchmark(self):
        df_bench, msg = self.data_api.daily(self.benchmark,
                                            start_date=self.extended_start_date_d, end_date=self.end_date,
                                            adjust_mode=self.adjust_mode,
                                            fields='trade_date,symbol,close,vwap,volume,turnover')
        # TODO: we want more than just close price of benchmark
        df_bench = df_bench.set_index('trade_date').loc[:, ['close']]
        return df_bench

    # --------------------------------------------------------------------------------------------------------
    # Add/Remove Fields&Formulas
    def _add_field(self, field_name, is_quarterly=None):
        self.fields.append(field_name)
        if not self._is_predefined_field(field_name):
            if is_quarterly is None:
                raise ValueError("Field [{:s}] is not a predefined field, but no frequency information is provided.")
            if is_quarterly:
                self.custom_quarterly_fields.append(field_name)
            else:
                self.custom_daily_fields.append(field_name)

    def add_field(self, field_name, data_api=None):
        """
        Query and append new field to DataView.
        
        Parameters
        ----------
        data_api : BaseDataServer
        field_name : str
            Must be a known field name (which is given in documents).
        
        Returns
        -------
        bool
            whether add successfully.

        """
        if data_api is None:
            if self.data_api is None:
                print("Add field failed. No data_api available. Please specify one in parameter.")
                return False
        else:
            self.data_api = data_api
            
        if field_name in self.fields:
            print("Field name [{:s}] already exists.".format(field_name))
            return False

        if not self._is_predefined_field(field_name):
            print("Field name [{}] not valid, ignore.".format(field_name))
            return False

        merge_d, merge_q = self._prepare_daily_quarterly([field_name])
    
        if self._is_daily_field(field_name):
            if self.data_d is None:
                raise ValueError("Please prepare [{:s}] first.".format(field_name))
            merge, _ = self._prepare_daily_quarterly([field_name])
            is_quarterly = False
        else:
            if self.data_q is None:
                raise ValueError("Please prepare [{:s}] first.".format(field_name))
            _, merge = self._prepare_daily_quarterly([field_name])
            is_quarterly = True
        
        merge = merge.loc[:, pd.IndexSlice[:, field_name]]
        merge.columns = merge.columns.droplevel(level=1)
        self.append_df(merge, field_name, is_quarterly=is_quarterly)  # whether contain only trade days is decided by existing data.
        
        if is_quarterly:
            df_ann = merge_q.loc[:, pd.IndexSlice[:, self.ANN_DATE_FIELD_NAME]]
            df_ann.columns = df_ann.columns.droplevel(level='field')
            df_expanded = align(merge, df_ann, self.dates)
            self.append_df(df_expanded, field_name, is_quarterly=False)
        return True
    
    def add_formula(self, field_name, formula, is_quarterly, overwrite=True,
                    formula_func_name_style='camel', data_api=None,
                    within_index=True):
        """
        Add a new field, which is calculated using existing fields.
        
        Parameters
        ----------
        formula : str or unicode
            A formula contains operations and function calls.
        field_name : str or unicode
            A custom name for the new field.
        is_quarterly : bool
            Whether df is quarterly data (like quarterly financial statement) or daily data.
        overwrite : bool, optional
            Whether overwrite existing field. True by default.
        formula_func_name_style : {'upper', 'lower'}, optional
        data_api : RemoteDataService, optional
        within_index : bool
            When do cross-section operatioins, whether just do within index components.
        
        Notes
        -----
        Time cost of this function:
            For a simple formula (like 'a + 1'), almost all time is consumed by append_df;
            For a complex formula (like 'GroupRank'), half of time is consumed by evaluation and half by append_df.
        """
        if data_api is not None:
            self.data_api = data_api
            
        if field_name in self.fields:
            if overwrite:
                self.remove_field(field_name)
                print("Field [{:s}] is overwritten.".format(field_name))
            else:
                raise ValueError("Add formula failed: name [{:s}] exist. Try another name.".format(field_name))
        elif self._is_predefined_field(field_name):
            raise ValueError("[{:s}] is alread a pre-defined field. Please use another name.".format(field_name))
        
        parser = Parser()
        parser.set_capital(formula_func_name_style)
        
        expr = parser.parse(formula)
        
        var_df_dic = dict()
        var_list = expr.variables()
        
        # TODO: users do not need to prepare data before add_formula
        if not self.fields:
            self.fields.extend(var_list)
            self.prepare_data()
        else:
            for var in var_list:
                if var not in self.fields:
                    print("Variable [{:s}] is not recognized (it may be wrong)," \
                          "try to fetch from the server...".format(var))
                    success = self.add_field(var)
                    if not success:
                        return
        
        for var in var_list:
            if self._is_quarter_field(var):
                df_var = self.get_ts_quarter(var, start_date=self.extended_start_date_q)
            else:
                # must use extended date. Default is start_date
                df_var = self.get_ts(var, start_date=self.extended_start_date_d, end_date=self.end_date)
            
            var_df_dic[var] = df_var
        
        # TODO: send ann_date into expr.evaluate. We assume that ann_date of all fields of a symbol is the same
        df_ann = self._get_ann_df()
        if within_index:
            df_index_member = self.get_ts('index_member', start_date=self.extended_start_date_d, end_date=self.end_date)
            df_eval = parser.evaluate(var_df_dic, ann_dts=df_ann, trade_dts=self.dates, index_member=df_index_member)
        else:
            df_eval = parser.evaluate(var_df_dic, ann_dts=df_ann, trade_dts=self.dates)

        self.append_df(df_eval, field_name, is_quarterly=is_quarterly)

        if is_quarterly:
            df_ann = self._get_ann_df()
            df_expanded = align(df_eval, df_ann, self.dates)
            self.append_df(df_expanded, field_name, is_quarterly=False)
        
    def append_df(self, df, field_name, is_quarterly=False):
        """
        Append DataFrame to existing multi-index DataFrame and add corresponding field name.
        
        Parameters
        ----------
        df : pd.DataFrame or pd.Series
        field_name : str or unicode
        is_quarterly : bool
            Whether df is quarterly data (like quarterly financial statement) or daily data.
            
        Notes
        -----
        append_df does not support overwrite. To overwrite a field, you must first do self.remove_fields(),
        then append_df() again.
        
        """
        df = df.copy()
        if isinstance(df, pd.DataFrame):
            pass
        elif isinstance(df, pd.Series):
            df = pd.DataFrame(df)
        else:
            raise ValueError("Data to be appended must be pandas format. But we have {}".format(type(df)))
    
        if is_quarterly:
            the_data = self.data_q
        else:
            the_data = self.data_d
        
        exist_symbols = the_data.columns.levels[0]
        if len(df.columns) < len(exist_symbols):
            df2 = pd.DataFrame(index=df.index, columns=exist_symbols, data=np.nan)
            df2.update(df)
            df = df2
        elif len(df.columns) > len(exist_symbols):
            df = df.loc[:, exist_symbols]
        multi_idx = pd.MultiIndex.from_product([exist_symbols, [field_name]])
        df.columns = multi_idx

        #the_data = apply_in_subprocess(pd.merge, args=(the_data, df),
        #                            kwargs={'left_index': True, 'right_index': True, 'how': 'left'})  # runs in *only* one process
        the_data = pd.merge(the_data, df, left_index=True, right_index=True, how='left')
        the_data = the_data.sort_index(axis=1)
        #merge = the_data.join(df, how='left')  # left: keep index of existing data unchanged
        #sort_columns(the_data)
    
        if is_quarterly:
            self.data_q = the_data
        else:
            self.data_d = the_data
        self._add_field(field_name, is_quarterly)

    def remove_field(self, field_names):
        """
        Query and append new field to DataView.
        
        Parameters
        ----------
        field_names : str
            Separated by comma.
            The (custom) field to be removed from dataview.
        
        Returns
        -------
        bool
            whether add successfully.

        """
        if isinstance(field_names, basestring):
            field_names = field_names.split(',')
        else:
            raise ValueError("field_names must be str separated by comma.")
    
        for field_name in field_names:
            # parameter validation
            if field_name not in self.fields:
                print("Field name [{:s}] does not exist. Stop remove_field.".format(field_name))
                return
        
            if self._is_daily_field(field_name):
                is_quarterly = False
            elif self._is_quarter_field(field_name):
                is_quarterly = True
            else:
                print("Field name [{}] is a pre-defined field, ignore.".format(field_name))
                return
        
            # remove field data
            
            self.data_d = self.data_d.drop(field_name, axis=1, level=1)
            if is_quarterly:
                self.data_q = self.data_q.drop(field_name, axis=1, level=1)
        
            # remove fields name from list
            self.fields.remove(field_name)
            if is_quarterly:
                if field_name in self.custom_quarterly_fields:
                    self.custom_quarterly_fields.remove(field_name)
            else:
                if field_name in self.custom_daily_fields:
                    self.custom_daily_fields.remove(field_name)

    # --------------------------------------------------------------------------------------------------------
    # Get Data API
    def get(self, symbol="", start_date=0, end_date=0, fields=""):
        """
        Basic API to get arbitrary data. If nothing fetched, return None.
        
        Parameters
        ----------
        symbol : str, optional
            Separated by ',' default "" (all securities).
        start_date : int, optional
            Default 0 (self.start_date).
        end_date : int, optional
            Default 0 (self.start_date).
        fields : str, optional
            Separated by ',' default "" (all fields).

        Returns
        -------
        res : pd.DataFrame or None
            index is datetimeindex, columns are (symbol, fields) MultiIndex

        """
        sep = ','
    
        if not fields:
            fields = slice(None)  # self.fields
        else:
            fields = fields.split(sep)
    
        if not symbol:
            symbol = slice(None)  # this is 3X faster than symbol = self.symbol
        else:
            symbol = symbol.split(sep)
    
        if not start_date:
            start_date = self.start_date
        if not end_date:
            end_date = self.end_date
    
        res = self.data_d.loc[pd.IndexSlice[start_date: end_date], pd.IndexSlice[symbol, fields]]
        return res
    
    def get_snapshot(self, snapshot_date, symbol="", fields=""):
        """
        Get snapshot of given fields and symbol at snapshot_date.
        
        Parameters
        ----------
        snapshot_date : int
            Date of snapshot.
        symbol : str, optional
            Separated by ',' default "" (all securities).
        fields : str, optional
            Separated by ',' default "" (all fields).

        Returns
        -------
        res : pd.DataFrame
            symbol as index, field as columns

        """

        # df = self.data_d.T.unstack()
        # df = df[snapshot_date].copy()
        if snapshot_date not in self._snapshot:
            return

        df = self._snapshot[snapshot_date]
        if fields:
            return df[fields.split(',')]
        else:
            return df


        res = self.get(symbol=symbol, start_date=snapshot_date, end_date=snapshot_date, fields=fields)
        if res is None:
            print("No data. for date={}, fields={}, symbol={}".format(snapshot_date, fields, symbol))
            return
    
        res = res.stack(level='symbol', dropna=False)
        res.index = res.index.droplevel(level=self.TRADE_DATE_FIELD_NAME)
    
        return res
    
    def _get_ann_df(self):
        """
        Query announcement date of financial statements of all securities.
        
        Returns
        -------
        df_ann : pd.DataFrame or None
            Index is date, column is symbol.
            If no quarterly data available, return None.
        
        """
        if self.data_q is None:
            return None
        df_ann = self.data_q.loc[:, pd.IndexSlice[:, self.ANN_DATE_FIELD_NAME]]
        df_ann.columns = df_ann.columns.droplevel(level='field')
    
        return df_ann

    def get_symbol(self, symbol, start_date=0, end_date=0, fields=""):
        res = self.get(symbol, start_date=start_date, end_date=end_date, fields=fields)
        if res is None:
            raise ValueError("No data. for "
                             "start_date={}, end_date={}, field={}, symbol={}".format(start_date, end_date,
                                                                                      fields, symbol))

        res.columns = res.columns.droplevel(level='symbol')
        return res

    def get_ts_quarter(self, field, symbol="", start_date=0, end_date=0):
        # TODO
        sep = ','
        if not symbol:
            symbol = self.symbol
        else:
            symbol = symbol.split(sep)
    
        if not start_date:
            start_date = self.start_date
        if not end_date:
            end_date = self.end_date
    
        df_ref_quarterly = self.data_q.loc[:, pd.IndexSlice[symbol, field]]
        df_ref_quarterly.columns = df_ref_quarterly.columns.droplevel(level='field')
    
        return df_ref_quarterly
    
    def get_ts(self, field, symbol="", start_date=0, end_date=0):
        """
        Get time series data of single field.
        
        Parameters
        ----------
        field : str or unicode
            Single field.
        symbol : str, optional
            Separated by ',' default "" (all securities).
        start_date : int, optional
            Default 0 (self.start_date).
        end_date : int, optional
            Default 0 (self.start_date).

        Returns
        -------
        res : pd.DataFrame
            Index is int date, column is symbol.

        """
        res = self.get(symbol, start_date=start_date, end_date=end_date, fields=field)
        if res is None:
            print("No data. for start_date={}, end_date={}, field={}, symbol={}".format(start_date,
                                                                                        end_date, field, symbol))
            raise ValueError
            return
    
        res.columns = res.columns.droplevel(level='field')
    
        return res
    
    # --------------------------------------------------------------------------------------------------------
    # DataView I/O
    @staticmethod
    def _load_h5(fp):
        """Load data and meta_data from hd5 file.
        
        Parameters
        ----------
        fp : str, optional
            File path of pre-stored hd5 file.
        
        """
        h5 = pd.HDFStore(fp)
        
        res = dict()
        for key in h5.keys():
            res[key] = h5.get(key)
            
        h5.close()
        
        return res
        
    def load_dataview(self, folder_path='.'):
        """
        Load data from local file.
        
        Parameters
        ----------
        folder_path : str or unicode, optional
            Folder path to store hd5 file and meta data.
            
        """
        path_meta_data = os.path.join(folder_path, 'meta_data.json')
        path_data = os.path.join(folder_path, 'data.hd5')
        if not (os.path.exists(path_meta_data) and os.path.exists(path_data)):
            raise IOError("There is no data file under directory {}".format(folder_path))
        
        meta_data = jutil.read_json(path_meta_data)
        dic = self._load_h5(path_data)
        self.data_d = dic.get('/data_d', None)
        self.data_q = dic.get('/data_q', None)
        self._data_benchmark = dic.get('/data_benchmark', None)
        self._data_inst = dic.get('/data_inst', None)
        self.__dict__.update(meta_data)

        a = self.get_ts('adjust_factor')
        b = (a / a.shift(1)).fillna(1.0)
        self.append_df(b, '_daily_adjust_factor', is_quarterly=False)

        dates = self.dates
        mask = dates < self.start_date
        before_first_day = dates[mask][-1]

        open = self.get_ts('open')
        preclose = self.get_ts('close', start_date=before_first_day).shift(1)
        limit = np.abs((open - preclose)/preclose)
        self.append_df(limit, "_limit", is_quarterly=False)

        dates = self.data_d.index.values
        df = self.data_d.T.unstack()
        self._snapshot = {}
        for date in dates:
            tmp = df[date].copy()
            del tmp.index.name
            del tmp.columns.name
            self._snapshot[date] = tmp

        print("Dataview loaded successfully.")

    def save_dataview(self, folder_path):
        """
        Save data and meta_data_to_store to a single hd5 file.
        Store at output/sub_folder
        
        Parameters
        ----------
        folder_path : str or unicode
            Path to store your data.

        """
        abs_folder = os.path.abspath(folder_path)
        meta_path = os.path.join(folder_path, 'meta_data.json')
        data_path = os.path.join(folder_path, 'data.hd5')
        
        data_to_store = {'data_d': self.data_d, 'data_q': self.data_q,
                         'data_benchmark': self.data_benchmark, 'data_inst': self.data_inst}
        data_to_store = {k: v for k, v in data_to_store.items() if v is not None}
        meta_data_to_store = {key: self.__dict__[key] for key in self.meta_data_list}

        print("\nStore data...")
        jutil.save_json(meta_data_to_store, meta_path)
        self._save_h5(data_path, data_to_store)
        
        print ("Dataview has been successfully saved to:\n"
               + abs_folder + "\n\n"
               + "You can load it with load_dataview('{:s}')".format(abs_folder))

    @staticmethod
    def _save_h5(fp, dic):
        """
        Save data in dic to a hd5 file.
        
        Parameters
        ----------
        fp : str
            File path.
        dic : dict

        """
        import warnings
        warnings.filterwarnings('ignore', category=pd.io.pytables.PerformanceWarning)
        
        jutil.create_dir(fp)
        h5 = pd.HDFStore(fp, complevel=9, complib='blosc')
        for key, value in dic.items():
            h5[key] = value
        h5.close()


class EventDataView(object):
    """
    Prepare data before research / trade. Support file I/O.
    Support: add field, add formula, save / load.
    
    Attributes
    ----------
    data_api : RemoteDataService
    symbol : list
    start_date : int
    end_date : int
    fields : list
    freq : int
    market_daily_fields, reference_daily_fields : list
    data_d : pd.DataFrame
        All daily frequency data will be merged and stored here.
        index is date, columns is symbol-field MultiIndex
    data_q : pd.DataFrame
        All quarterly frequency data will be merged and stored here.
        index is date, columns is symbol-field MultiIndex
    
    """
    def __init__(self):
        self.data_api = None
        
        self.universe = ""
        self.symbol = []
        self.benchmark = ""
        self.start_date = 0
        self.extended_start_date_d = 0
        self.end_date = 0
        self.fields = []
        self.freq = 1
        self.all_price = True
        
        self.meta_data_list = ['start_date', 'end_date',
                               'extended_start_date_d',
                               'freq', 'fields', 'symbol', 'universe', 'benchmark',
                               'custom_daily_fields']
        self.adjust_mode = 'post'
        
        self.data_d = None
        self.data_q = None
        self._data_benchmark = None
        self._data_inst = None
        self.data_custom = None
        # self._data_group = None
        
        common_list = {'symbol', 'start_date', 'end_date'}
        market_bar_list = {'open', 'high', 'low', 'close', 'volume', 'turnover', 'vwap', 'oi'}
        market_tick_list = {'volume', 'oi',
                            'askprice1', 'askprice2', 'askprice3', 'askprice4', 'askprice5',
                            'bidprice1', 'bidprice1', 'bidprice1', 'bidprice1', 'bidprice1',
                            'askvolume1', 'askvolume2', 'askvolume3', 'askvolume4', 'askvolume5',
                            'bidvolume1', 'bidvolume2', 'bidvolume3', 'bidvolume4', 'bidvolume5'}
        # fields map
        # TODO: 'freq' is not in market_daily_fields yet.
        self.market_daily_fields = \
            {'open', 'high', 'low', 'close', 'volume', 'turnover', 'vwap', 'oi', 'trade_status',
             'open_adj', 'high_adj', 'low_adj', 'close_adj', 'vwap_adj', 'index_member', 'index_weight'}
        self.group_fields = {'sw1', 'sw2', 'sw3', 'sw4', 'zz1', 'zz2'}
        self .custom_daily_fields = []
        
        # const
        self.TRADE_STATUS_FIELD_NAME = 'trade_status'
        self.TRADE_DATE_FIELD_NAME = 'trade_date'
    
    # --------------------------------------------------------------------------------------------------------
    # Properties
    @property
    def data_benchmark(self):
        return self._data_benchmark
    
    @property
    def data_inst(self):
        """
        
        Returns
        -------
        pd.DataFrame

        """
        return self._data_inst
    
    @data_benchmark.setter
    def data_benchmark(self, df_new):
        if self.data_d is not None and df_new.shape[0] != self.data_d.shape[0]:
            raise ValueError("You must provide a DataFrame with the same shape of data_benchmark.")
        self._data_benchmark = df_new
    
    @property
    def dates(self):
        """
        Get daily date array of the underlying data.
        
        Returns
        -------
        res : np.array
            dtype: int

        """
        if self.data_d is not None:
            res = self.data_d.index.values
        elif self.data_api is not None:
            res = self.data_api.query_trade_dates(self.extended_start_date_d, self.end_date)
        else:
            raise ValueError("Cannot get dates array when neither of data and data_api exists.")
        
        return res
    
    # --------------------------------------------------------------------------------------------------------
    # Fields
    def _is_quarter_field(self, field_name):
        """
        Check whether a field name is quarterly frequency.
        field_name must be pre-defined or already added.
        
        Parameters
        ----------
        field_name : str

        Returns
        -------
        bool

        """
        res = (field_name in self.fin_stat_balance_sheet
               or field_name in self.fin_stat_cash_flow
               or field_name in self.fin_stat_income
               or field_name in self.fin_indicator
               or field_name in self.custom_quarterly_fields)
        return res
    
    def _is_daily_field(self, field_name):
        """
        Check whether a field name is daily frequency.
        field_name must be pre-defined or already added.
        
        Parameters
        ----------
        field_name : str

        Returns
        -------
        bool

        """
        flag = (field_name in self.market_daily_fields
                or field_name in self.reference_daily_fields
                or field_name in self.custom_daily_fields
                or field_name in self.group_fields)
        return flag
    
    def _is_predefined_field(self, field_name):
        """
        Check whether a field name can be recognized.
        field_name must be pre-defined or already added.
        
        Parameters
        ----------
        field_name : str

        Returns
        -------
        bool

        """
        return self._is_quarter_field(field_name) or self._is_daily_field(field_name)
    
    def _get_fields(self, field_type, fields, complement=False, append=False):
        """
        Get list of fields that are in ref_quarterly_fields.
        
        Parameters
        ----------
        field_type : {'market_daily', 'ref_daily', 'income', 'balance_sheet', 'cash_flow', 'daily', 'quarterly'
        fields : list of str
        complement : bool, optional
            If True, get fields that are NOT in ref_quarterly_fields.

        Returns
        -------
        list

        """
        pool_map = {'market_daily': self.market_daily_fields,
                    'group': self.group_fields}
        pool_map['daily'] = set.union(pool_map['market_daily'],
                                      pool_map['group'],
                                      self.custom_daily_fields)
        
        pool = pool_map.get(field_type, None)
        if pool is None:
            raise NotImplementedError("field_type = {:s}".format(field_type))
        
        s = set.intersection(set(pool), set(fields))
        if not s:
            return []
        
        if complement:
            s = set(fields) - s
        
        if field_type == 'market_daily' and self.all_price:
            # turnover will not be adjusted
            s.update({'open', 'high', 'close', 'low', 'vwap'})
        
        if append:
            s.add('symbol')
            if field_type == 'market_daily' or field_type == 'ref_daily':
                s.add('trade_date')
                if field_type == 'market_daily':
                    s.add(self.TRADE_STATUS_FIELD_NAME)
            elif (field_type == 'income'
                  or field_type == 'balance_sheet'
                  or field_type == 'cash_flow'
                  or field_type == 'fin_indicator'):
                s.add(self.ANN_DATE_FIELD_NAME)
                s.add(self.REPORT_DATE_FIELD_NAME)
        
        l = list(s)
        return l
    
    # --------------------------------------------------------------------------------------------------------
    # Prepare data
    def init_from_config(self, props, data_api):
        """
        Initialize various attributes like start/end date, universe/symbol, fields, etc.
        If your want to parse symbol, but use a custom benchmark index, please directly assign self.data_benchmark.
        
        Parameters
        ----------
        props : dict
            start_date, end_date, freq, symbol, fields, etc.
        data_api : BaseDataServer
        
        """
        # data_api.init_from_config(props)
        self.data_api = data_api
        
        sep = ','
        
        # initialize parameters
        self.start_date = props['start_date']
        self.extended_start_date_d = jutil.shift(self.start_date, n_weeks=-8)  # query more data
        self.extended_start_date_q = jutil.shift(self.start_date, n_weeks=-130)
        self.end_date = props['end_date']
        self.all_price = props.get('all_price', True)
        self.freq = props.get('freq', 1)
        
        # get and filter fields
        fields = props.get('fields', [])
        if fields:
            fields = props['fields'].split(sep)
            self.fields = [field for field in fields if self._is_predefined_field(field)]
            if len(self.fields) < len(fields):
                print("Field name [{}] not valid, ignore.".format(set.difference(set(fields), set(self.fields))))
        
        # append additional fields
        if self.all_price:
            self.fields.extend(['open_adj', 'high_adj', 'low_adj', 'close_adj',
                                'open', 'high', 'low', 'close',
                                'vwap', 'vwap_adj'])
        
        # initialize universe/symbol
        universe = props.get('universe', "")
        symbol = props.get('symbol', "")
        benchmark = props.get('benchmark', '')
        if symbol and universe:
            raise ValueError("Please use either [symbol] or [universe].")
        if not (symbol or universe):
            raise ValueError("One of [symbol] or [universe] must be provided.")
        if universe:
            univ_list = universe.split(',')
            self.universe = univ_list
            symbols_list = []
            for univ in univ_list:
                symbols_list.extend(data_api.query_index_member(univ, self.extended_start_date_d, self.end_date))
            self.symbol = sorted(list(set(symbols_list)))
        else:
            self.symbol = sorted(symbol.split(sep))
        if benchmark:
            self.benchmark = benchmark
        else:
            if self.universe:
                if len(self.universe) > 1:
                    print("More than one universe are used: {}, "
                          "use the first one ({}) as index by default. "
                          "If you want to use other benchmark, "
                          "please specify benchmark in configs.".format(repr(self.universe), self.universe[0]))
                self.benchmark = self.universe[0]
        
        print("Initialize config success.")

    def distributed_query(self, query_func_name, symbol, start_date, end_date, limit=100000, **kwargs):
        n_symbols = len(symbol.split(','))
        dates = self.data_api.query_trade_dates(start_date, end_date)
        n_days = len(dates)
    
        if n_symbols * n_days > limit:
            n = limit // n_symbols
        
            df_list = []
            i = 0
            pos1, pos2 = n * i, n * (i + 1) - 1
            while pos2 < n_days:
                print(pos2)
                df, msg = getattr(self.data_api, query_func_name)(symbol=symbol,
                                                                  start_date=dates[pos1], end_date=dates[pos2],
                                                                  **kwargs)
                df_list.append(df)
                i += 1
                pos1, pos2 = n * i, n * (i + 1) - 1
            if pos1 < n_days:
                df, msg = getattr(self.data_api, query_func_name)(symbol=symbol,
                                                                  start_date=dates[pos1], end_date=dates[-1],
                                                                  **kwargs)
                df_list.append(df)
            df = pd.concat(df_list, axis=0)
        else:
            df, msg = getattr(self.data_api, query_func_name)(symbol,
                                                              start_date=start_date, end_date=end_date,
                                                              **kwargs)
        return df, msg
    
    def prepare_data(self):
        """Prepare data for the FIRST time."""
        # prepare benchmark and group
        print("Query data...")
        data_d = self._prepare_daily_quarterly(self.fields)
        self.data_d = data_d
        #self._align_and_merge_q_into_d()
        
        print("Query instrument info...")
        self._prepare_inst_info()
        
        if self.benchmark:
            print("Query benchmark...")
            self._data_benchmark = self._prepare_benchmark()
        
        print("Data has been successfully prepared.")
    
    @staticmethod
    def _process_index_co(df, index_name):
        df = df.astype(dtype={index_name: int})
        df = df.drop_duplicates(subset=['symbol', index_name])
        return df
    
    def _prepare_daily_quarterly(self, fields):
        """
        Query and process data from data_api.
        
        Parameters
        ----------
        fields : list

        Returns
        -------
        merge_d : pd.DataFrame or None
        merge_q : pd.DataFrame or None

        """
        if not fields:
            return None, None
        
        # query data
        print("Query data - query...")
        daily_list = self._query_data(self.symbol, fields)
        
        def pivot_and_sort(df, index_name):
            df = self._process_index_co(df, index_name)
            df = df.pivot(index=index_name, columns='symbol')
            df.columns = df.columns.swaplevel()
            col_names = ['symbol', 'field']
            df.columns.names = col_names
            df = df.sort_index(axis=1, level=col_names)
            df.index.name = index_name
            return df
        
        multi_daily = None
        if daily_list:
            daily_list_pivot = [pivot_and_sort(df, self.TRADE_DATE_FIELD_NAME) for df in daily_list]
            multi_daily = self._merge_data(daily_list_pivot, self.TRADE_DATE_FIELD_NAME)
            # use self.dates as index because original data have weekends
            multi_daily = self._fill_missing_idx_col(multi_daily, index=self.dates, symbols=self.symbol)
            print("Query data - daily fields prepared.")
        
        return multi_daily
    
    def _query_data(self, symbol, fields):
        """
        Query data using different APIs, then store them in dict.
        period, start_date and end_date are fixed.
        Keys of dict are securitites.
        
        Parameters
        ----------
        symbol : list of str
        fields : list of str

        Returns
        -------
        daily_list : list
        quarterly_list : list

        """
        sep = ','
        symbol_str = sep.join(symbol)
        
        if self.freq == 1:
            daily_list = []
            
            # TODO : use fields = {field: kwargs} to enable params
            fields_market_daily = self._get_fields('market_daily', fields, append=True)
            if fields_market_daily:
                print("NOTE: price adjust method is [{:s} adjust]".format(self.adjust_mode))
                # no adjust prices and other market daily fields
                df_daily, msg1 = self.distributed_query('daily', symbol_str,
                                                        start_date=self.extended_start_date_d, end_date=self.end_date,
                                                        adjust_mode=None, fields=sep.join(fields_market_daily), limit=100000)
                #df_daily, msg1 = self.data_api.daily(symbol_str, start_date=self.extended_start_date_d, end_date=self.end_date,
                #                                     adjust_mode=None, fields=sep.join(fields_market_daily))
                
                if self.all_price:
                    adj_cols = ['open', 'high', 'low', 'close', 'vwap']
                    # adjusted prices
                    #df_daily_adjust, msg11 = self.data_api.daily(symbol_str, start_date=self.extended_start_date_d, end_date=self.end_date,
                    #                                             adjust_mode=self.adjust_mode, fields=','.join(adj_cols))
                    df_daily_adjust, msg1 = self.distributed_query('daily', symbol_str,
                                                                   start_date=self.extended_start_date_d, end_date=self.end_date,
                                                                   adjust_mode=self.adjust_mode, fields=sep.join(fields_market_daily), limit=100000)
                    
                    df_daily = pd.merge(df_daily, df_daily_adjust, how='outer',
                                        on=['symbol', 'trade_date'], suffixes=('', '_adj'))
                daily_list.append(df_daily.loc[:, fields_market_daily])
        
        else:
            raise NotImplementedError("freq = {}".format(self.freq))
        
        return daily_list
    
    @staticmethod
    def _merge_data(dfs, index_name='trade_date'):
        """
        Merge data from different APIs into one DataFrame.
        
        Parameters
        ----------
        dfs : list of pd.DataFrame

        Returns
        -------
        merge : pd.DataFrame or None
            If dfs is empty, return None
        
        Notes
        -----
        Align on date index, concatenate on columns (symbol and fields)
        
        """
        # dfs = [df for df in dfs if df is not None]
        
        merge = pd.concat(dfs, axis=1, join='outer')
        
        # drop duplicated columns. ONE LINE EFFICIENT version
        mask_duplicated = merge.columns.duplicated()
        if np.any(mask_duplicated):
            # print("Duplicated columns found. Dropped.")
            merge = merge.loc[:, ~mask_duplicated]
            
            # if merge.isnull().sum().sum() > 0:
            # print "WARNING: nan in final merged data. NO fill"
            # merge.fillna(method='ffill', inplace=True)
        
        merge = merge.sort_index(axis=1, level=['symbol', 'field'])
        merge.index.name = index_name
        
        return merge
    
    def _fill_missing_idx_col(self, df, index=None, symbols=None):
        if index is None:
            index = df.index
        if symbols is None:
            symbols = self.symbol
        fields = df.columns.levels[1]
        
        if len(fields) * len(self.symbol) != len(df.columns) or len(index) != len(df.index):
            cols_multi = pd.MultiIndex.from_product([symbols, fields], names=['symbol', 'field'])
            cols_multi = cols_multi.sort_values()
            df_final = pd.DataFrame(index=index, columns=cols_multi, data=np.nan)
            df_final.index.name = df.index.name
            
            df_final.update(df)
            
            # idx_diff = sorted(set(df_final.index) - set(df.index))
            col_diff = sorted(set(df_final.columns.levels[0].values) - set(df.columns.levels[0].values))
            print ("WARNING: some data is unavailable: "
                   # + "\n    At index " + ', '.join(idx_diff)
                   + "\n    At fields " + ', '.join(col_diff))
            return df_final
        else:
            return df
    
    def _prepare_adj_factor(self):
        """Query and append daily adjust factor for prices."""
        mask_stocks = self.data_inst['inst_type'] == 1
        if mask_stocks.sum() == 0:
            return
        symbol_stocks = self.data_inst.loc[mask_stocks].index.values
        symbol_str = ','.join(symbol_stocks)
        df_adj = self.data_api.query_adj_factor_daily(symbol_str,
                                                      start_date=self.extended_start_date_d, end_date=self.end_date, div=False)
        self.append_df(df_adj, 'adjust_factor', is_quarterly=False)
    
    def _prepare_comp_info(self):
        # if a symbol is index member of any one universe, its value of index_member will be 1.0
        res = dict()
        for univ in self.universe:
            df = self.data_api.query_index_member_daily(univ, self.extended_start_date_d, self.end_date)
            res[univ] = df
        df_res = pd.concat(res, axis=0)
        df = df_res.groupby(by='trade_date').apply(lambda df: df.any(axis=0)).astype(float)
        self.append_df(df, 'index_member', is_quarterly=False)
        
        # use weights of the first universe
        df_weights = self.data_api.query_index_weights_daily(self.universe[0], self.extended_start_date_d, self.end_date)
        self.append_df(df_weights, 'index_weight', is_quarterly=False)
    
    def _prepare_inst_info(self):
        res = self.data_api.query_inst_info(symbol=','.join(self.symbol),
                                            fields='symbol,inst_type,name,list_date,'
                                                   'delist_date,product,pricetick,multiplier,'
                                                   'buylot,setlot',
                                            inst_type="")
        self._data_inst = res
    
    def _prepare_group(self, group_fields):
        data_map = {'sw1': ('SW', 1),
                    'sw2': ('SW', 2),
                    'sw3': ('SW', 3),
                    'sw4': ('SW', 4),
                    'zz1': ('ZZ', 1),
                    'zz2': ('ZZ', 2)}
        for field in group_fields:
            type_, level = data_map[field]
            df = self.data_api.query_industry_daily(symbol=','.join(self.symbol),
                                                    start_date=self.extended_start_date_q, end_date=self.end_date,
                                                    type_=type_, level=level)
            self.append_df(df, field, is_quarterly=False)
    
    def _prepare_benchmark(self):
        df_bench, msg = self.data_api.daily(self.benchmark,
                                            start_date=self.extended_start_date_d, end_date=self.end_date,
                                            adjust_mode=self.adjust_mode,
                                            fields='trade_date,symbol,close,vwap,volume,turnover')
        # TODO: we want more than just close price of benchmark
        df_bench = df_bench.set_index('trade_date').loc[:, ['close']]
        return df_bench
    
    # --------------------------------------------------------------------------------------------------------
    # Add/Remove Fields&Formulas
    def _add_field(self, field_name, is_quarterly=None):
        self.fields.append(field_name)
        if not self._is_predefined_field(field_name):
            self.custom_daily_fields.append(field_name)

    def _add_symbol(self, symbol_name):
        if symbol_name in self.symbol:
            print("symbol [{:s}] already exists, add_symbol failed.".format(symbol_name))
            return
        self.symbol.append(symbol_name)

    def add_field(self, field_name, data_api=None):
        """
        Query and append new field to DataView.
        
        Parameters
        ----------
        data_api : BaseDataServer
        field_name : str
            Must be a known field name (which is given in documents).
        
        Returns
        -------
        bool
            whether add successfully.

        """
        if data_api is None:
            if self.data_api is None:
                print("Add field failed. No data_api available. Please specify one in parameter.")
                return False
        else:
            self.data_api = data_api
        
        if field_name in self.fields:
            print("Field name [{:s}] already exists.".format(field_name))
            return False
        
        if not self._is_predefined_field(field_name):
            print("Field name [{}] not valid, ignore.".format(field_name))
            return False
        
        merge_d, merge_q = self._prepare_daily_quarterly([field_name])
        
        if self._is_daily_field(field_name):
            if self.data_d is None:
                raise ValueError("Please prepare [{:s}] first.".format(field_name))
            merge, _ = self._prepare_daily_quarterly([field_name])
            is_quarterly = False
        else:
            if self.data_q is None:
                raise ValueError("Please prepare [{:s}] first.".format(field_name))
            _, merge = self._prepare_daily_quarterly([field_name])
            is_quarterly = True
        
        merge = merge.loc[:, pd.IndexSlice[:, field_name]]
        merge.columns = merge.columns.droplevel(level=1)
        self.append_df(merge, field_name, is_quarterly=is_quarterly)  # whether contain only trade days is decided by existing data.
        
        if is_quarterly:
            df_ann = merge_q.loc[:, pd.IndexSlice[:, self.ANN_DATE_FIELD_NAME]]
            df_ann.columns = df_ann.columns.droplevel(level='field')
            df_expanded = align(merge, df_ann, self.dates)
            self.append_df(df_expanded, field_name, is_quarterly=False)
        return True
    
    def add_formula(self, field_name, formula, is_quarterly, overwrite=True,
                    formula_func_name_style='camel', data_api=None,
                    within_index=True):
        """
        Add a new field, which is calculated using existing fields.
        
        Parameters
        ----------
        formula : str
            A formula contains operations and function calls.
        field_name : str
            A custom name for the new field.
        is_quarterly : bool
            Whether df is quarterly data (like quarterly financial statement) or daily data.
        overwrite : bool, optional
            Whether overwrite existing field. True by default.
        formula_func_name_style : {'upper', 'lower'}, optional
        data_api : RemoteDataService, optional
        within_index : bool
            When do cross-section operatioins, whether just do within index components.
        
        Notes
        -----
        Time cost of this function:
            For a simple formula (like 'a + 1'), almost all time is consumed by append_df;
            For a complex formula (like 'GroupRank'), half of time is consumed by evaluation and half by append_df.
        """
        if data_api is not None:
            self.data_api = data_api
        
        if field_name in self.fields:
            if overwrite:
                self.remove_field(field_name)
                print("Field [{:s}] is overwritten.".format(field_name))
            else:
                print("Add formula failed: name [{:s}] exist. Try another name.".format(field_name))
                return
        
        parser = Parser()
        parser.set_capital(formula_func_name_style)
        
        expr = parser.parse(formula)
        
        var_df_dic = dict()
        var_list = expr.variables()
        
        # TODO: users do not need to prepare data before add_formula
        if not self.fields:
            self.fields.extend(var_list)
            self.prepare_data()
        else:
            for var in var_list:
                if var not in self.fields:
                    print("Variable [{:s}] is not recognized (it may be wrong)," \
                          "try to fetch from the server...".format(var))
                    success = self.add_field(var)
                    if not success:
                        return
        
        for var in var_list:
            if self._is_quarter_field(var):
                df_var = self.get_ts_quarter(var, start_date=self.extended_start_date_q)
            else:
                # must use extended date. Default is start_date
                df_var = self.get_ts(var, start_date=self.extended_start_date_d, end_date=self.end_date)
            
            var_df_dic[var] = df_var
        
        # TODO: send ann_date into expr.evaluate. We assume that ann_date of all fields of a symbol is the same
        df_ann = self._get_ann_df()
        if within_index:
            df_index_member = self.get_ts('index_member', start_date=self.extended_start_date_d, end_date=self.end_date)
            df_eval = parser.evaluate(var_df_dic, ann_dts=df_ann, trade_dts=self.dates, index_member=df_index_member)
        else:
            df_eval = parser.evaluate(var_df_dic, ann_dts=df_ann, trade_dts=self.dates)
        
        self.append_df(df_eval, field_name, is_quarterly=is_quarterly)
        
        if is_quarterly:
            df_ann = self._get_ann_df()
            df_expanded = align(df_eval, df_ann, self.dates)
            self.append_df(df_expanded, field_name, is_quarterly=False)
    
    def append_df(self, df, field_name, is_quarterly=False):
        """
        Append DataFrame to existing multi-index DataFrame and add corresponding field name.
        
        Parameters
        ----------
        df : pd.DataFrame or pd.Series
        field_name : str
        is_quarterly : bool
            Whether df is quarterly data (like quarterly financial statement) or daily data.
            
        Notes
        -----
        append_df does not support overwrite. To overwrite a field, you must first do self.remove_fields(),
        then append_df() again.
        
        """
        df = df.copy()
        if isinstance(df, pd.DataFrame):
            pass
        elif isinstance(df, pd.Series):
            df = pd.DataFrame(df)
        else:
            raise ValueError("Data to be appended must be pandas format. But we have {}".format(type(df)))
        
        if is_quarterly:
            the_data = self.data_q
        else:
            the_data = self.data_d
        
        exist_symbols = the_data.columns.levels[0]
        if len(df.columns) < len(exist_symbols):
            df2 = pd.DataFrame(index=df.index, columns=exist_symbols, data=np.nan)
            df2.update(df)
            df = df2
        elif len(df.columns) > len(exist_symbols):
            df = df.loc[:, exist_symbols]
        multi_idx = pd.MultiIndex.from_product([exist_symbols, [field_name]])
        df.columns = multi_idx
        
        #the_data = apply_in_subprocess(pd.merge, args=(the_data, df),
        #                            kwargs={'left_index': True, 'right_index': True, 'how': 'left'})  # runs in *only* one process
        the_data = pd.merge(the_data, df, left_index=True, right_index=True, how='left')
        the_data = the_data.sort_index(axis=1)
        #merge = the_data.join(df, how='left')  # left: keep index of existing data unchanged
        #sort_columns(the_data)
        
        if is_quarterly:
            self.data_q = the_data
        else:
            self.data_d = the_data
        self._add_field(field_name, is_quarterly)

    def append_df_symbol(self, df, symbol_name):
        """
        Append DataFrame to existing multi-index DataFrame and add corresponding field name.
        
        Parameters
        ----------
        df : pd.DataFrame or pd.Series
        symbol_name : str
        is_quarterly : bool
            Whether df is quarterly data (like quarterly financial statement) or daily data.
            
        Notes
        -----
        append_df does not support overwrite. To overwrite a field, you must first do self.remove_fields(),
        then append_df() again.
        
        """
        df = df.copy()
        if isinstance(df, pd.DataFrame):
            pass
        elif isinstance(df, pd.Series):
            df = pd.DataFrame(df)
        else:
            raise ValueError("Data to be appended must be pandas format. But we have {}".format(type(df)))
    
        the_data = self.data_d
    
        exist_fields = the_data.columns.levels[1]
        if len(set(exist_fields) - set(df.columns)):
        #if set(df.columns) < set(exist_fields):
            df2 = pd.DataFrame(index=df.index, columns=exist_fields, data=np.nan)
            df2.update(df)
            df = df2
        multi_idx = pd.MultiIndex.from_product([[symbol_name], exist_fields])
        df.columns = multi_idx
    
        #the_data = apply_in_subprocess(pd.merge, args=(the_data, df),
        #                            kwargs={'left_index': True, 'right_index': True, 'how': 'left'})  # runs in *only* one process
        the_data = pd.merge(the_data, df, left_index=True, right_index=True, how='left')
        the_data = the_data.sort_index(axis=1)
        #merge = the_data.join(df, how='left')  # left: keep index of existing data unchanged
        #sort_columns(the_data)
    
        self.data_d = the_data
        self._add_symbol(symbol_name)

    def remove_field(self, field_names):
        """
        Query and append new field to DataView.
        
        Parameters
        ----------
        field_names : str or list
            The (custom) field to be removed from dataview.
        
        Returns
        -------
        bool
            whether add successfully.

        """
        if isinstance(field_names, basestring):
            field_names = field_names.split(',')
        elif isinstance(field_names, (list, tuple)):
            pass
        else:
            raise ValueError("field_names must be str or list of str.")
        
        for field_name in field_names:
            # parameter validation
            if field_name not in self.fields:
                print("Field name [{:s}] does not exist.".format(field_name))
                return
            
            if self._is_daily_field(field_name):
                is_quarterly = False
            elif self._is_quarter_field(field_name):
                is_quarterly = True
            else:
                print("Field name [{}] is a pre-defined field, ignore.".format(field_name))
                return
            
            # remove field data
            
            self.data_d = self.data_d.drop(field_name, axis=1, level=1)
            if is_quarterly:
                self.data_q = self.data_q.drop(field_name, axis=1, level=1)
            
            # remove fields name from list
            self.fields.remove(field_name)
            if is_quarterly:
                if field_name in self.custom_quarterly_fields:
                    self.custom_quarterly_fields.remove(field_name)
            else:
                if field_name in self.custom_daily_fields:
                    self.custom_daily_fields.remove(field_name)
    
    # --------------------------------------------------------------------------------------------------------
    # Get Data API
    def get(self, symbol="", start_date=0, end_date=0, fields="", data_format='wide'):
        """
        Basic API to get arbitrary data. If nothing fetched, return None.
        
        Parameters
        ----------
        symbol : str, optional
            Separated by ',' default "" (all securities).
        start_date : int, optional
            Default 0 (self.start_date).
        end_date : int, optional
            Default 0 (self.start_date).
        fields : str, optional
            Separated by ',' default "" (all fields).
        data_format : {'long', 'wide'}, optional
            Format of result DataFrame, default 'wide'.

        Returns
        -------
        res : pd.DataFrame or None
            index is datetimeindex, columns are (symbol, fields) MultiIndex

        """
        sep = ','
        
        if not fields:
            fields = slice(None)  # self.fields
        else:
            fields = fields.split(sep)
        
        if not symbol:
            symbol = slice(None)  # this is 3X faster than symbol = self.symbol
        else:
            symbol = symbol.split(sep)
        
        if not start_date:
            start_date = self.start_date
        if not end_date:
            end_date = self.end_date
        
        res = self.data_d.loc[pd.IndexSlice[start_date: end_date], pd.IndexSlice[symbol, fields]]
        
        if data_format == 'wide':
            pass
        else:
            res = res.stack(level='symbol').reset_index()
        return res
    
    def get_snapshot(self, snapshot_date, symbol="", fields=""):
        """
        Get snapshot of given fields and symbol at snapshot_date.
        
        Parameters
        ----------
        snapshot_date : int
            Date of snapshot.
        symbol : str, optional
            Separated by ',' default "" (all securities).
        fields : str, optional
            Separated by ',' default "" (all fields).

        Returns
        -------
        res : pd.DataFrame
            symbol as index, field as columns

        """
        res = self.get(symbol=symbol, start_date=snapshot_date, end_date=snapshot_date, fields=fields)
        if res is None:
            print("No data. for date={}, fields={}, symbol={}".format(snapshot_date, fields, symbol))
            return
        
        res = res.stack(level='symbol', dropna=False)
        res.index = res.index.droplevel(level=self.TRADE_DATE_FIELD_NAME)
        
        return res
    
    def _get_ann_df(self):
        """
        Query announcement date of financial statements of all securities.
        
        Returns
        -------
        df_ann : pd.DataFrame or None
            Index is date, column is symbol.
            If no quarterly data available, return None.
        
        """
        if self.data_q is None:
            return None
        df_ann = self.data_q.loc[:, pd.IndexSlice[:, self.ANN_DATE_FIELD_NAME]]
        df_ann.columns = df_ann.columns.droplevel(level='field')
        
        return df_ann
    
    def get_symbol(self, symbol, start_date=0, end_date=0, fields=""):
        res = self.get(symbol, start_date=start_date, end_date=end_date, fields=fields)
        if res is None:
            raise ValueError("No data. for "
                             "start_date={}, end_date={}, field={}, symbol={}".format(start_date, end_date,
                                                                                      fields, symbol))

        res.columns = res.columns.droplevel(level='symbol')
        return res

    def get_ts(self, field, symbol="", start_date=0, end_date=0):
        """
        Get time series data of single field.
        
        Parameters
        ----------
        field : str
            Single field.
        symbol : str, optional
            Separated by ',' default "" (all securities).
        start_date : int, optional
            Default 0 (self.start_date).
        end_date : int, optional
            Default 0 (self.start_date).

        Returns
        -------
        res : pd.DataFrame
            Index is int date, column is symbol.

        """
        res = self.get(symbol, start_date=start_date, end_date=end_date, fields=field)
        if res is None:
            raise ValueError("No data. for "
                             "start_date={}, end_date={}, field={}, symbol={}".format(start_date, end_date,
                                                                                      field, symbol))
        
        res.columns = res.columns.droplevel(level='field')
        return res
    
    # --------------------------------------------------------------------------------------------------------
    # DataView I/O
    @staticmethod
    def _load_h5(fp):
        """Load data and meta_data from hd5 file.
        
        Parameters
        ----------
        fp : str, optional
            File path of pre-stored hd5 file.
        
        """
        h5 = pd.HDFStore(fp)
        
        res = dict()
        for key in h5.keys():
            res[key] = h5.get(key)
        
        h5.close()
        
        return res
    
    def load_dataview(self, folder_path='.'):
        """
        Load data from local file.
        
        Parameters
        ----------
        folder_path : str, optional
            Folder path to store hd5 file and meta data.
            
        """
        path_meta_data = os.path.join(folder_path, 'meta_data.json')
        path_data = os.path.join(folder_path, 'data.hd5')
        if not (os.path.exists(path_meta_data) and os.path.exists(path_data)):
            raise IOError("There is no data file under directory {}".format(folder_path))
        
        meta_data = jutil.read_json(path_meta_data)
        dic = self._load_h5(path_data)
        self.data_d = dic.get('/data_d', None)
        self.data_q = dic.get('/data_q', None)
        self._data_benchmark = dic.get('/data_benchmark', None)
        self._data_inst = dic.get('/data_inst', None)
        self.data_custom = dic.get('/data_custom', None)
        self.__dict__.update(meta_data)
        
        print("Dataview loaded successfully.")
    
    def save_dataview(self, folder_path):
        """
        Save data and meta_data_to_store to a single hd5 file.
        Store at output/sub_folder
        
        Parameters
        ----------
        folder_path : str
            Path to store your data.

        """
        abs_folder = os.path.abspath(folder_path)
        meta_path = os.path.join(folder_path, 'meta_data.json')
        data_path = os.path.join(folder_path, 'data.hd5')
        
        data_to_store = {'data_d': self.data_d, 'data_q': self.data_q,
                         'data_benchmark': self.data_benchmark, 'data_inst': self.data_inst,
                         'data_custom': self.data_custom}
        data_to_store = {k: v for k, v in data_to_store.items() if v is not None}
        meta_data_to_store = {key: self.__dict__[key] for key in self.meta_data_list}
        
        print("\nStore data...")
        jutil.save_json(meta_data_to_store, meta_path)
        self._save_h5(data_path, data_to_store)
        
        print ("Dataview has been successfully saved to:\n"
               + abs_folder + "\n\n"
               + "You can load it with load_dataview('{:s}')".format(abs_folder))
    
    @staticmethod
    def _save_h5(fp, dic):
        """
        Save data in dic to a hd5 file.
        
        Parameters
        ----------
        fp : str
            File path.
        dic : dict

        """
        import warnings
        warnings.filterwarnings('ignore', category=pd.io.pytables.PerformanceWarning)
        
        jutil.create_dir(fp)
        h5 = pd.HDFStore(fp, complevel=9, complib='blosc')
        for key, value in dic.items():
            h5[key] = value
        h5.close()




# encoding: utf-8
"""
dataapi defines standard APIs for communicating with data service.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .data_api import DataApi

__all__ = ['DataApi']



# encoding: UTF-8
"""
The event processing engine is where the event is identified,
and the appropriate reaction is selected and executed.

Our framework utilizes event engine to run in an efficient way.

"""
from .engine import EventEngine, Event
from .eventtype import EVENT_TYPE



# encoding: utf-8
"""
tradeapi defines standard APIs for communicating with
algorithmic trading & executing system.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .trade_api import TradeApi

__all__ = ['TradeApi']



# encoding:utf-8
"""
Classes defined in order module are relevant to orders, including:
- Order
- OrderStatusInd
- Task

"""

from __future__ import print_function
from jaqs.trade import common


class Order(object):
    """
    Basic order class.

    Attributes
    ----------
    ba_id : str
        Id of broker account.
    sa_id : str
        Id of strategy account.
    task_id : str
        Id of the task (which may contain several orders).
    entrust_no : str
        Id of the order.
    symbol : str
    entrust_action : str
        Action of the order.
    entrust_price : float
        Target price of the order.
    entrust_size : int
        Target quantity of the order.
    entrust_date : int
        Placement date of the order.
    entrust_time : int
        Placement time of the order.
    ord_seq : int
        Number of sub-orders, start with 0.
    batch_no : int
        Number of batch.
    order_status : str
        Current status of the order.
    fill_price : float
        Average fill price of the order.
    fill_size : int
        Average fill size of the order.
    algo : str
        Algorithm of the task.
    order_type : str (common.ORDER_TYPE)
        market, limit, stop, etc.
    time_in_force : str (common.ORDER_TIME_IN_FORCE)
        FAK, FOK, GTM, etc.

    Methods
    -------
    copy

    """
    def __init__(self, order=None):
        """If order is provided, copy all attributes of it."""
        self.ba_id = ""
        self.sa_id = ""
        self.task_id = 0
        self.entrust_no = ""
        
        self.symbol = ""
        
        self.entrust_action = ""
        self.entrust_price = 0.0
        self.entrust_size = 0
        self.entrust_date = 0
        self.entrust_time = 0
        
        self.ord_seq = 0
        self.batch_no = 0
        
        self.order_status = ""
        self.fill_price = 0.0
        self.fill_size = 0
        
        self.algo = ""
        
        self.order_type = ""
        self.time_in_force = ""
        
        self.commission = 0.0
        
        if order is not None:
            self.copy(order)
        
    def __repr__(self):
        return "{0.entrust_date:8d}({0.entrust_time:8d}) " \
               "{0.entrust_action:6s} {0.symbol:10s}@{0.entrust_price:.3f}" \
               "  size = {0.fill_size} / {0.entrust_size}".format(self)

    def __str__(self):
        return self.__repr__()

    def copy(self, order):
        """
        Copy all attributes of order.
        
        Parameters
        ----------
        order : Order

        """
        self.task_id = order.task_id
        self.entrust_no = order.entrust_no
        
        self.symbol = order.symbol
        
        self.entrust_action = order.entrust_action
        self.entrust_price = order.entrust_price
        self.entrust_size = order.entrust_size
        self.entrust_date = order.entrust_date
        self.entrust_time = order.entrust_time
        
        self.ord_seq = order.ord_seq
        self.batch_no = order.batch_no
        
        self.order_status = order.order_status
        self.fill_size = order.fill_size
        self.fill_price = order.fill_price
        
        self.algo = order.algo
        
        self.order_type = order.order_type
        self.time_in_force = order.time_in_force
        
        self.commission = order.commission
    
    @property
    def is_finished(self):
        return (self.order_status == common.ORDER_STATUS.FILLED
                or self.order_status == common.ORDER_STATUS.CANCELLED
                or self.order_status == common.ORDER_STATUS.REJECTED)
    
    @classmethod
    def new_order(cls, symbol, action, price, size, date, time, order_type=None):
        """Create a new order based on attributes provided and return it."""
        o = cls()
        o.symbol = symbol
        o.entrust_action = action
        o.entrust_price = price
        o.entrust_size = size
        o.entrust_date = date
        o.entrust_time = time
        o.order_status = common.ORDER_STATUS.NEW
        if order_type is None:
            o.order_type = common.ORDER_TYPE.LIMIT
        else:
            o.order_type = order_type
        
        return o


class FixedPriceTypeOrder(Order):
    """
    This type of order aims to be matched at a given price type, eg. CLOSE, OPEN, VWAP, etc.
    Only used in daily resolution trade.

    Attributes
    ----------
    price_target : str
        The type of price we want.

    Methods
    -------

    """
    
    def __init__(self, target=""):
        Order.__init__(self)
        
        self.price_target = target
    

class VwapOrder(Order):
    """
    This type of order will only be matched once a day.
    Only used in daily resolution trade.

    Attributes
    ----------
    start : int
        The start of matching time range.
        If start = -1, end will be ignored and the order will be matched in the whole trading session.
    end : int
        The end of matching time range.

    """
    
    def __init__(self, start=-1, end=-1):
        Order.__init__(self)
        
        self.start = start
        self.end = end
    
    @property
    def time_range(self):
        return self.start, self.end


class OrderStatusInd(object):
    """
    OrderStatusInd is a indication of status change of an order.
    
    Attributes
    ----------
    ba_id : str
        Id of broker account.
    sa_id : str
        Id of strategy account.
    task_id : str
        Id of the task (which may contain several orders).
    entrust_no : str
        Id of the order.
    symbol : str
    entrust_action : str
        Action of the order.
    entrust_price : float
        Target price of the order.
    entrust_size : int
        Target quantity of the order.
    entrust_date : int
        Placement date of the order.
    entrust_time : int
        Placement time of the order.
    ord_seq : int
        Number of sub-orders, start with 0.
    batch_no : int
        Number of batch.
    order_status : str
        Current status of the order.
    fill_price : float
        Average fill price of the order.
    fill_size : int
        Average fill size of the order.
    algo : str
        Algorithm of the task.
    order_type : str (common.ORDER_TYPE)
        market, limit, stop, etc.
    time_in_force : str (common.ORDER_TIME_IN_FORCE)
        FAK, FOK, GTM, etc.
        
    """
    def __init__(self, order=None):
        self.ba_id = 0
        self.sa_id = 0
        
        self.task_id = 0
        self.entrust_no = ''
        
        self.order_type = ""
        self.algo = ''
        self.batch_no = 0
        self.ord_seq = 0
        
        self.symbol = ''
        
        self.entrust_action = ''
        self.entrust_price = 0.0
        self.entrust_size = 0
        self.entrust_date = 0
        self.entrust_time = 0
        
        self.order_status = ''
        
        self.fill_size = 0
        self.fill_price = 0.0
        self.commission = 0.0
        
        self.time_in_force = ""
        
        if order is not None:
            if isinstance(order, Order):
                self.init_from_order(order)
            else:
                raise ValueError("init_from_order only accept argument of type Order.")
    
    def init_from_order(self, order):
        """Update information from a given order."""
        self.entrust_no = order.entrust_no
        self.task_id = order.task_id
        
        self.symbol = order.symbol
        
        self.entrust_action = order.entrust_action
        self.entrust_price = order.entrust_price
        self.entrust_size = order.entrust_size
        self.entrust_date = order.entrust_date
        self.entrust_time = order.entrust_time
        
        self.order_status = order.order_status
        
        self.fill_size = order.fill_size
        self.fill_price = order.fill_price

        self.algo = order.algo
        self.order_type = order.order_type

    @classmethod
    def create_from_dict(cls, dic):
        ind = cls()
        ind.__dict__.update(dic)
        return ind
    
    def __repr__(self):
        return "{0.order_status:8s}  |  " \
               "{0.entrust_date:8d}({0.entrust_time:8d}) " \
               "{0.entrust_action:6s} {0.symbol:10s}@{0.entrust_price:.3f}" \
               "  size = {0.entrust_size}".format(self)

    def __str__(self):
        return self.__repr__()


'''
class OrderRsp(object):
    def __init__(self, entrust_no="", msg="", task_id=0):
        self.msg = msg
        self.entrust_no = entrust_no
        self.task_id = task_id

    @classmethod
    def create_from_dict(cls, dic):
        rsp = cls()
        rsp.__dict__.update(dic)
        return rsp
    
    def __repr__(self):
        return "entrust_no = {0.entrust_no:s}  |  task_id = {0.task_id:d}" \
               "\nmsg = '{0.msg:s}'".format(self)

    def __str__(self):
        return self.__repr__()

'''

'''
class TaskRsp(object):
    def __init__(self, task_no=0, task_id=0, msg=""):
        self.msg = msg
        self.task_no = task_no
        self.task_id = task_id
    
    @classmethod
    def create_from_dict(cls, dic):
        rsp = cls()
        rsp.__dict__.update(dic)
        return rsp
    
    def __repr__(self):
        return "task_no = {0.task_no:s}  |  task_id = {0.task_id:d}" \
               "\nmsg = '{0.msg:s}'".format(self)
    
    def __str__(self):
        return self.__repr__()
    
    @property
    def success(self):
        return (self.task_id is not None) and (self.task_id != 0)

'''


class Task(object):
    """
    Task is a high-level trading target, which may contain many orders.
    
    Attributes
    ----------
    task_id : int
        Id of the task.
    data : Order, list of Order, goals, etc.
        Order information of this task.
    algo : str
        Algorithm of the task.
    algo_param : str
        Parameters of algorithm of the task.
    trade_date : int
        %YY%mm%dd
    function_name : str
        The function that generated this task. Used to identify task type.
    
    
    """
    def __init__(self, task_id=0,
                 algo="", algo_param=None,
                 data=None,
                 function_name="", trade_date=0):
        self.task_id = task_id
        self.task_status = common.TASK_STATUS.ACCEPTED
        
        self.data = data
        
        self.algo = algo
        self.algo_param = dict() if algo_param else algo_param
        
        self.trade_date = trade_date
        self.function_name = function_name
    
    @property
    def is_finished(self):
        return self.task_status == common.TASK_STATUS.DONE



# encoding: utf-8
from jaqs.util.sequence import SequenceGenerator


def test_seq_gen():
    sg = SequenceGenerator()
    for i in range(1, 999):
        assert sg.get_next('order') == i

    text = 'trade'
    sg.get_next(text)
    sg.get_next(text)
    for i in range(3, 999):
        assert sg.get_next(text) == i


if __name__ == "__main__":
    test_seq_gen()



# encoding: utf-8

from __future__ import print_function
from jaqs.data import RemoteDataService
from jaqs.data import DataView
import jaqs.util as jutil

from config_path import DATA_CONFIG_PATH
data_config = jutil.read_json(DATA_CONFIG_PATH)

daily_path = '../output/tests/test_dataview_d'
quarterly_path = '../output/tests/test_dataview_q'


def test_write():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv = DataView()
    
    secs = '600030.SH,000063.SZ,000001.SZ'
    props = {'start_date': 20160601, 'end_date': 20170601, 'symbol': secs,
             'fields': 'open,close,high,low,volume,pb,net_assets,pcf_ncf',
             'freq': 1}

    dv.init_from_config(props, data_api=ds)
    dv.prepare_data()
    assert dv.data_d.shape == (281, 48)
    assert dv.dates.shape == (281, )
    # TODO
    """
    PerformanceWarning:
    your performance may suffer as PyTables will pickle object types that it cannot
    map directly to c-types [inferred_type->mixed,key->block1_values] [items->[('000001.SZ', 'int_income'), ('000001.SZ', 'less_handling_chrg_comm_exp'), ('000001.SZ', 'net_int_income'), ('000001.SZ', 'oper_exp'), ('000001.SZ', 'symbol'), ('000063.SZ', 'int_income'), ('000063.SZ', 'less_handling_chrg_comm_exp'), ('000063.SZ', 'net_int_income'), ('000063.SZ', 'oper_exp'), ('000063.SZ', 'symbol'), ('600030.SH', 'int_income'), ('600030.SH', 'less_handling_chrg_comm_exp'), ('600030.SH', 'net_int_income'), ('600030.SH', 'oper_exp'), ('600030.SH', 'symbol')]]
    """
    
    dv.save_dataview(folder_path=daily_path)


def test_write_future():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv = DataView()
    
    secs = 'rb1710.SHF,j1710.DCE'
    props = {'start_date': 20170401, 'end_date': 20170901, 'symbol': secs,
             'fields': 'open,close,high,low,volume,oi',
             'freq': 1, 'all_price': False}
    
    dv.init_from_config(props, data_api=ds)
    dv.prepare_data()
    assert dv.data_d.shape == (145, 14)
    
    
def test_load():
    dv = DataView()
    dv.load_dataview(folder_path=daily_path)
    
    assert dv.start_date == 20160601 and set(dv.symbol) == set('000001.SZ,600030.SH,000063.SZ'.split(','))

    # test get_snapshot
    snap1 = dv.get_snapshot(20170504, symbol='600030.SH,000063.SZ', fields='close,pb')
    assert snap1.shape == (2, 2)
    assert set(snap1.columns.values) == {'close', 'pb'}
    assert set(snap1.index.values) == {'600030.SH', '000063.SZ'}
    
    # test get_ts
    ts1 = dv.get_ts('close', symbol='600030.SH,000063.SZ', start_date=20170101, end_date=20170302)
    assert ts1.shape == (38, 2)
    assert set(ts1.columns.values) == {'600030.SH', '000063.SZ'}
    assert ts1.index.values[-1] == 20170302


def test_add_field():
    dv = DataView()
    dv.load_dataview(folder_path=daily_path)
    nrows, ncols = dv.data_d.shape
    n_securities = len(dv.data_d.columns.levels[0])
    
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv.add_field('total_share', ds)
    assert dv.data_d.shape == (nrows, ncols + 1 * n_securities)


def test_add_formula_directly():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv = DataView()
    
    secs = '600030.SH,000063.SZ,000001.SZ'
    props = {'start_date': 20160601, 'end_date': 20170601, 'symbol': secs,
             'fields': 'open,close',
             'freq': 1}
    dv.init_from_config(props, data_api=ds)
    dv.prepare_data()
    
    dv.add_formula("myfactor", 'close / open', is_quarterly=False)
    assert dv.data_d.shape == (281, 39)


def test_add_formula():
    dv = DataView()
    dv.load_dataview(folder_path=daily_path)
    nrows, ncols = dv.data_d.shape
    n_securities = len(dv.data_d.columns.levels[0])
    
    formula = 'Delta(high - close, 1)'
    dv.add_formula('myvar1', formula, is_quarterly=False)
    assert dv.data_d.shape == (nrows, ncols + 1 * n_securities)
    
    formula2 = 'myvar1 - close'
    dv.add_formula('myvar2', formula2, is_quarterly=False)
    assert dv.data_d.shape == (nrows, ncols + 2 * n_securities)


def test_dataview_universe():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv = DataView()
    
    props = {'start_date': 20170227, 'end_date': 20170327, 'universe': '000016.SH',
             # 'symbol': 'rb1710.SHF,rb1801.SHF',
             'fields': ('open,high,low,close,vwap,volume,turnover,'
                        + 'sw1,zz2,'
                        + 'roe,net_assets,'
                        + 'total_oper_rev,oper_exp,tot_profit,int_income'
                        ),
             'freq': 1}
    
    dv.init_from_config(props, ds)
    dv.prepare_data()
    
    data_bench = dv.data_benchmark.copy()
    dv.data_benchmark = data_bench
    
    try:
        dv.data_benchmark = data_bench.iloc[3:]
    except ValueError:
        pass
    
    dv.remove_field('roe,net_assets')
    dv.remove_field('close')


# quarterly
def test_q():
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv = DataView()
    
    secs = '600030.SH,000063.SZ,000001.SZ'
    props = {'start_date': 20160609, 'end_date': 20170601, 'symbol': secs,
             'fields': ('open,close,'
                        + 'pb,net_assets,'
                        + 'total_oper_rev,oper_exp,'
                        + 'cash_paid_invest,'
                        + 'capital_stk,'
                        + 'roe'), 'freq': 1}
    
    dv.init_from_config(props, data_api=ds)
    dv.prepare_data()
    dv.save_dataview(folder_path=quarterly_path)


def test_q_get():
    dv = DataView()
    dv.load_dataview(folder_path=quarterly_path)
    res = dv.get("", 0, 0, 'total_oper_rev')
    assert set(res.index.values) == set(dv.dates[dv.dates >= dv.start_date])


def test_q_add_field():
    dv = DataView()
    dv.load_dataview(folder_path=quarterly_path)
    nrows, ncols = dv.data_q.shape
    n_securities = len(dv.data_d.columns.levels[0])
    
    ds = RemoteDataService()
    ds.init_from_config(data_config)
    dv.add_field('net_inc_other_ops', ds)
    """
    dv.add_field('oper_rev', ds)
    dv.add_field('turnover', ds)
    """
    assert dv.data_q.shape == (nrows, ncols + 1 * n_securities)


def test_q_add_formula():
    dv = DataView()
    folder_path = '../output/prepared/20160609_20170601_freq=1D'
    dv.load_dataview(folder_path=quarterly_path)
    nrows, ncols = dv.data_d.shape
    n_securities = len(dv.data_d.columns.levels[0])
    
    formula = 'total_oper_rev / close'
    dv.add_formula('myvar1', formula, is_quarterly=False)
    df1 = dv.get_ts('myvar1')
    assert not df1.empty
    
    formula2 = 'Delta(oper_exp * myvar1 - open, 3)'
    dv.add_formula('myvar2', formula2, is_quarterly=False)
    df2 = dv.get_ts('myvar2')
    assert not df2.empty


if __name__ == "__main__":
    g = globals()
    g = {k: v for k, v in g.items() if k.startswith('test_') and callable(v)}

    # for test_name, test_func in g.items():
    for test_name in ['test_write', 'test_load', 'test_add_field', 'test_add_formula_directly',
                      'test_add_formula', 'test_dataview_universe',
                      'test_q', 'test_q_get', 'test_q_add_field', 'test_q_add_formula',
                      ]:
        test_func = g[test_name]
        print("\n==========\nTesting {:s}...".format(test_name))
        test_func()
    print("Test Complete.")



# encoding: utf-8

from __future__ import print_function
from functools import wraps

import numpy as np
import pandas as pd

import matplotlib as mpl
mpl.use('Agg')
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter
import matplotlib.gridspec as gridspec
import seaborn as sns

from . import performance as pfm
import jaqs.util as jutil


DECIMAL_TO_BPS = 10000
DECIMAL_TO_PCT = 100
COLOR_MAP = cm.get_cmap('rainbow') # cm.get_cmap('RdBu')
MPL_RCPARAMS = {'figure.facecolor': '#F6F6F6',
                'axes.facecolor': '#F6F6F6',
                'axes.edgecolor': '#D3D3D3',
                'text.color': '#555555',
                'grid.color': '#B1B1B1',
                'grid.alpha': 0.3,
                # scale
                'axes.linewidth': 2.0,
                'axes.titlepad': 12,
                'grid.linewidth': 1.0,
                'grid.linestyle': '-',
                # font size
                'font.size': 13,
                'axes.titlesize': 18,
                'axes.labelsize': 14,
                'legend.fontsize': 'small',
                'lines.linewidth': 2.5,
                }
mpl.rcParams.update(MPL_RCPARAMS)

# -----------------------------------------------------------------------------------
# plotting settings


def customize(func):
    """
    Decorator to set plotting context and axes style during function call.
    """
    @wraps(func)
    def call_w_context(*args, **kwargs):
        set_context = kwargs.pop('set_context', True)
        if set_context:
            with plotting_context(), axes_style():
                sns.despine(left=True)
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return call_w_context


def plotting_context(context='notebook', font_scale=1.5, rc=None):
    """
    Create signaldigger default plotting style context.

    Under the hood, calls and returns seaborn.plotting_context() with
    some custom settings. Usually you would use in a with-context.

    Parameters
    ----------
    context : str, optional
        Name of seaborn context.
    font_scale : float, optional
        Scale font by signal font_scale.
    rc : dict, optional
        Config flags.
        By default, {'lines.linewidth': 1.5}
        is being used and will be added to any
        rc passed in, unless explicitly overriden.

    Returns
    -------
    seaborn plotting context

    Example
    -------
    with signaldigger.plotting.plotting_context(font_scale=2):
        signaldigger.create_full_report(..., set_context=False)

    See also
    --------
    For more information, see seaborn.plotting_context().
    """
    if rc is None:
        rc = {}

    rc_default = {'lines.linewidth': 1.5}

    # Add defaults if they do not exist
    for name, val in rc_default.items():
        rc.setdefault(name, val)

    return sns.plotting_context(context=context, font_scale=font_scale, rc=rc)


def axes_style(style='darkgrid', rc=None):
    """Create signaldigger default axes style context.

    Under the hood, calls and returns seaborn.axes_style() with
    some custom settings. Usually you would use in a with-context.

    Parameters
    ----------
    style : str, optional
        Name of seaborn style.
    rc : dict, optional
        Config flags.

    Returns
    -------
    seaborn plotting context

    Example
    -------
    with signaldigger.plotting.axes_style(style='whitegrid'):
        signaldigger.create_full_report(..., set_context=False)

    See also
    --------
    For more information, see seaborn.plotting_context().

    """
    if rc is None:
        rc = {}

    rc_default = {}

    # Add defaults if they do not exist
    for name, val in rc_default.items():
        rc.setdefault(name, val)

    return sns.axes_style(style=style, rc=rc)


class GridFigure(object):
    def __init__(self, rows, cols, height_ratio=1.0):
        self.rows = rows * 2
        self.cols = cols
        self.fig = plt.figure(figsize=(14, rows * 7 * height_ratio))
        self.gs = gridspec.GridSpec(self.rows, self.cols, wspace=0.1, hspace=0.5)
        self.curr_row = 0
        self.curr_col = 0
        
        self._in_row = False
    
    def next_row(self):
        if self._in_row:
            self.curr_row += 2
            self.curr_col = 0
            self._in_row = False
        
        subplt = plt.subplot(self.gs[self.curr_row: self.curr_row + 2, :])
        self.curr_row += 2
        return subplt
    
    def next_subrow(self):
        if self._in_row:
            self.curr_row += 2
            self.curr_col = 0
            self._in_row = False
        
        subplt = plt.subplot(self.gs[self.curr_row, :])
        self.curr_row += 1
        return subplt
    
    def next_cell(self):
        subplt = plt.subplot(self.gs[self.curr_row: self.curr_row + 2, self.curr_col])
        self.curr_col += 1
        self._in_row = True
        if self.curr_col >= self.cols:
            self.curr_row += 2
            self.curr_col = 0
            self._in_row = False
        return subplt


# -----------------------------------------------------------------------------------
# Functions to Plot Tables


def plot_table(table, name=None, fmt=None):
    """
    Pretty print a pandas DataFrame.

    Uses HTML output if running inside Jupyter Notebook, otherwise
    formatted text output.

    Parameters
    ----------
    table : pd.Series or pd.DataFrame
        Table to pretty-print.
    name : str, optional
        Table name to display in upper left corner.
    fmt : str, optional
        Formatter to use for displaying table elements.
        E.g. '{0:.2f}%' for displaying 100 as '100.00%'.
        Restores original setting after displaying.
    """
    if isinstance(table, pd.Series):
        table = pd.DataFrame(table)
    
    if isinstance(table, pd.DataFrame):
        table.columns.name = name
    
    prev_option = pd.get_option('display.float_format')
    if fmt is not None:
        pd.set_option('display.float_format', lambda x: fmt.format(x))
    
    print(table)
    
    if fmt is not None:
        pd.set_option('display.float_format', prev_option)


def plot_information_table(ic_summary_table):
    print("Information Analysis")
    plot_table(ic_summary_table.apply(lambda x: x.round(3)).T)


def plot_quantile_statistics_table(tb):
    print("\n\nValue of signals of Different Quantiles Statistics")
    plot_table(tb)


# -----------------------------------------------------------------------------------
# Functions to Plot Returns


'''


def plot_quantile_returns_bar(mean_ret_by_q,
                              # ylim_percentiles=None,
                              ax=None):
    """
    Plots mean period wise returns for signal quantiles.

    Parameters
    ----------
    mean_ret_by_q : pd.DataFrame
        DataFrame with quantile, (group) and mean period wise return values.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
        
    """
    mean_ret_by_q = mean_ret_by_q.copy().loc[:, ['mean']]
    
    ymin = None
    ymax = None
    
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))
    
    mean_ret_by_q.multiply(DECIMAL_TO_BPS) \
        .plot(kind='bar',
              title="Mean Return (on symbol, time) By signal Quantile", ax=ax)
    ax.set(xlabel='Quantile', ylabel='Mean Return (bps)',
           ylim=(ymin, ymax))
    
    return ax


'''


def plot_quantile_returns_ts(mean_ret_by_q, ax=None):
    """
    Plots mean period wise returns for signal quantiles.

    Parameters
    ----------
    mean_ret_by_q : pd.DataFrame
        DataFrame with quantile, (group) and mean period wise return values.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
        
    """
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))
    
    ret_wide = pd.concat({k: v['mean'] for k, v in mean_ret_by_q.items()}, axis=1)
    ret_wide.index = pd.to_datetime(ret_wide.index, format="%Y%m%d")
    ret_wide = ret_wide.mul(DECIMAL_TO_PCT)
    # ret_wide = ret_wide.rolling(window=22).mean()
    
    ret_wide.plot(lw=1.2, ax=ax, cmap=COLOR_MAP)
    df = pd.DataFrame()
    ax.legend(loc='upper left')
    ymin, ymax = ret_wide.min().min(), ret_wide.max().max()
    ax.set(ylabel='Return (%)',
           title="Daily Quantile Return (equal weight within quantile)",
           xlabel='Date',
           # yscale='symlog',
           # yticks=np.linspace(ymin, ymax, 5),
           ylim=(ymin, ymax))
    
    ax.yaxis.set_major_formatter(ScalarFormatter())
    ax.axhline(1.0, linestyle='-', color='black', lw=1)
    
    return ax


def plot_mean_quantile_returns_spread_time_series(mean_returns_spread, period,
                                                  std_err=None,
                                                  bandwidth=1,
                                                  ax=None):
    """
    Plots mean period wise returns for signal quantiles.

    Parameters
    ----------
    mean_returns_spread : pd.Series
        Series with difference between quantile mean returns by period.
    std_err : pd.Series
        Series with standard error of difference between quantile
        mean returns each period.
    bandwidth : float
        Width of displayed error bands in standard deviations.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    
    if False:  # isinstance(mean_returns_spread, pd.DataFrame):
        if ax is None:
            ax = [None for a in mean_returns_spread.columns]
        
        ymin, ymax = (None, None)
        for (i, a), (name, fr_column) in zip(enumerate(ax),
                                             mean_returns_spread.items()):
            stdn = None if std_err is None else std_err[name]
            stdn = mean_returns_spread.loc
            a = plot_mean_quantile_returns_spread_time_series(fr_column,
                                                              std_err=stdn,
                                                              ax=a)
            ax[i] = a
            curr_ymin, curr_ymax = a.get_ylim()
            ymin = curr_ymin if ymin is None else min(ymin, curr_ymin)
            ymax = curr_ymax if ymax is None else max(ymax, curr_ymax)
        
        for a in ax:
            a.set_ylim([ymin, ymax])
        
        return ax
    
    periods = period
    title = ('Top Minus Bottom Quantile Return'
             .format(periods if periods is not None else ""))
    
    if ax is None:
        f, ax = plt.subplots(figsize=(18, 6))
    
    mean_returns_spread.index = pd.to_datetime(mean_returns_spread.index, format="%Y%m%d")
    mean_returns_spread_bps = mean_returns_spread['mean_diff'] * DECIMAL_TO_PCT

    std_err_bps = mean_returns_spread['std'] * DECIMAL_TO_PCT
    upper = mean_returns_spread_bps.values + (std_err_bps * bandwidth)
    lower = mean_returns_spread_bps.values - (std_err_bps * bandwidth)
    
    mean_returns_spread_bps.plot(alpha=0.4, ax=ax, lw=0.7, color='navy')
    mean_returns_spread_bps.rolling(22).mean().plot(color='green',
                                                    alpha=0.7,
                                                    ax=ax)
    # ax.fill_between(mean_returns_spread.index, lower, upper,
    #                 alpha=0.3, color='indianred')
    ax.axhline(0.0, linestyle='-', color='black', lw=1, alpha=0.8)

    ax.legend(['mean returns spread', '1 month moving avg'], loc='upper right')
    ylim = np.nanpercentile(abs(mean_returns_spread_bps.values), 95)
    ax.set(ylabel='Difference In Quantile Mean Return (%)',
           xlabel='',
           title=title,
           ylim=(-ylim, ylim))
    
    return ax


def plot_cumulative_return(ret, ax=None, title=None):
    """
    Plots the cumulative returns of the returns series passed in.

    Parameters
    ----------
    ret : pd.Series
        Period wise returns of dollar neutral portfolio weighted by signal
        value.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))
    
    ret = ret.copy()
    
    cum = ret  # pfm.daily_ret_to_cum(ret)
    cum.index = pd.to_datetime(cum.index, format="%Y%m%d")
    cum = cum.mul(DECIMAL_TO_PCT)
    
    cum.plot(ax=ax, lw=3, color='indianred', alpha=1.0)
    ax.axhline(0.0, linestyle='-', color='black', lw=1)
    
    metrics = pfm.calc_performance_metrics(cum, cum_return=True, compound=False)
    ax.text(.85, .30,
            "Ann.Ret. = {:.1f}%\nAnn.Vol. = {:.1f}%\nSharpe = {:.2f}".format(metrics['ann_ret'],
                                                                           metrics['ann_vol'],
                                                                           metrics['sharpe']),
            fontsize=12,
            bbox={'facecolor': 'white', 'alpha': 1, 'pad': 5},
            transform=ax.transAxes,
            verticalalignment='top')
    if title is None:
        title = "Cumulative Return"
    ax.set(ylabel='Cumulative Return (%)',
           title=title,
           xlabel='Date')
    
    return ax


def plot_cumulative_returns_by_quantile(quantile_ret, ax=None):
    """
    Plots the cumulative returns of various signal quantiles.

    Parameters
    ----------
    quantile_ret : int: pd.DataFrame
        Cumulative returns by signal quantile.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
    """
    
    if ax is None:
        f, ax = plt.subplots(1, 1, figsize=(18, 6))
    
    cum_ret = quantile_ret
    cum_ret.index = pd.to_datetime(cum_ret.index, format="%Y%m%d")
    cum_ret = cum_ret.mul(DECIMAL_TO_PCT)
    
    cum_ret.plot(lw=2, ax=ax, cmap=COLOR_MAP)
    ax.axhline(0.0, linestyle='-', color='black', lw=1)
    
    ax.legend(loc='upper left')
    ymin, ymax = cum_ret.min().min(), cum_ret.max().max()
    ax.set(ylabel='Cumulative Returns (%)',
           title='Cumulative Return of Each Quantile (equal weight within quantile)',
           xlabel='Date',
           # yscale='symlog',
           # yticks=np.linspace(ymin, ymax, 5),
           ylim=(ymin, ymax))
    
    sharpes = ["sharpe_{:d} = {:.2f}".format(col, pfm.calc_performance_metrics(ser, cum_return=True,
                                                                               compound=False)['sharpe'])
               for col, ser in cum_ret.iteritems()]
    ax.text(.02, .30,
            '\n'.join(sharpes),
            fontsize=12,
            bbox={'facecolor': 'white', 'alpha': 1, 'pad': 5},
            transform=ax.transAxes,
            verticalalignment='top')

    ax.yaxis.set_major_formatter(ScalarFormatter())
    
    return ax


# -----------------------------------------------------------------------------------
# Functions to Plot IC


def plot_ic_ts(ic, period, ax=None):
    """
    Plots Spearman Rank Information Coefficient and IC moving
    average for a given signal.

    Parameters
    ----------
    ic : pd.DataFrame
        DataFrame indexed by date, with IC for each forward return.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    ic = ic.copy()
    if isinstance(ic, pd.DataFrame):
        ic = ic.iloc[:, 0]
    mean, std = ic.mean(), ic.std()

    if ax is None:
        num_plots = 1
        f, ax = plt.subplots(num_plots, 1, figsize=(18, num_plots * 7))
        ax = np.asarray([ax]).flatten()

    ic.plot(ax=ax, lw=0.6, color='navy', label='daily IC', alpha=0.8)
    ic.rolling(22).mean().plot(ax=ax, color='royalblue', lw=2, alpha=0.6, label='1 month MA')
    ax.axhline(0.0, linestyle='-', color='black', linewidth=1, alpha=0.8)

    ax.text(.05, .95,
            "Mean {:.3f} \n Std. {:.3f}".format(mean, std),
            fontsize=16,
            bbox={'facecolor': 'white', 'alpha': 1, 'pad': 5},
            transform=ax.transAxes,
            verticalalignment='top',
            )
    
    ymin, ymax = (None, None)
    curr_ymin, curr_ymax = ax.get_ylim()
    ymin = curr_ymin if ymin is None else min(ymin, curr_ymin)
    ymax = curr_ymax if ymax is None else max(ymax, curr_ymax)

    ax.legend(loc='upper right')
    ax.set(ylabel='IC', xlabel="", ylim=[ymin, ymax],
           title="Daily IC and Moving Average".format(period))
    
    return ax


def plot_ic_hist(ic, period, ax=None):
    """
    Plots Spearman Rank Information Coefficient histogram for a given signal.

    Parameters
    ----------
    ic : pd.DataFrame
        DataFrame indexed by date, with IC for each forward return.
    ax : matplotlib.Axes, optional
        Axes upon which to plot.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """

    ic = ic.copy()
    if isinstance(ic, pd.DataFrame):
        ic = ic.iloc[:, 0]
    mean, std = ic.mean(), ic.std()
    
    if ax is None:
        v_spaces = 1
        f, ax = plt.subplots(v_spaces, 3, figsize=(18, v_spaces * 6))
        ax = ax.flatten()

    sns.distplot(ic.replace(np.nan, 0.), ax=ax,
                 hist_kws={'color': 'royalblue'},
                 kde_kws={'color': 'navy', 'alpha': 0.5},
                 # hist_kws={'weights':},
                 )
    ax.axvline(mean, color='indianred', linestyle='dashed', linewidth=1.0, label='Mean')
    ax.text(.05, .95,
            "Mean {:.3f} \n Std. {:.3f}".format(mean, std),
            fontsize=16,
            bbox={'facecolor': 'white', 'alpha': 1, 'pad': 5},
            transform=ax.transAxes,
            verticalalignment='top')
    
    ax.set(title="Distribution of Daily IC",
           xlabel='IC',
           xlim=[-1, 1])
    ax.legend(loc='upper right')

    return ax


def plot_monthly_ic_heatmap(mean_monthly_ic, period, ax=None):
    """
    Plots a heatmap of the information coefficient or returns by month.

    Parameters
    ----------
    mean_monthly_ic : pd.DataFrame
        The mean monthly IC for N periods forward.

    Returns
    -------
    ax : matplotlib.Axes
        The axes that were plotted on.
    """
    MONTH_MAP = {1: 'Jan',
                 2: 'Feb',
                 3: 'Mar',
                 4: 'Apr',
                 5: 'May',
                 6: 'Jun',
                 7: 'Jul',
                 8: 'Aug',
                 9: 'Sep',
                 10: 'Oct',
                 11: 'Nov',
                 12: 'Dec'}
    
    mean_monthly_ic = mean_monthly_ic.copy()
    
    num_plots = 1.0
    
    v_spaces = ((num_plots - 1) // 3) + 1
    
    if ax is None:
        f, ax = plt.subplots(v_spaces, 3, figsize=(18, v_spaces * 6))
        ax = ax.flatten()
    
    new_index_year = []
    new_index_month = []
    for date in mean_monthly_ic.index:
        new_index_year.append(date.year)
        new_index_month.append(MONTH_MAP[date.month])
    
    mean_monthly_ic.index = pd.MultiIndex.from_arrays(
            [new_index_year, new_index_month],
            names=["year", "month"])
    
    ic_year_month = mean_monthly_ic['ic'].unstack()
    sns.heatmap(
            ic_year_month,
            annot=True,
            alpha=1.0,
            center=0.0,
            annot_kws={"size": 7},
            linewidths=0.01,
            linecolor='white',
            cmap=cm.get_cmap('RdBu'),
            cbar=False,
            ax=ax)
    ax.set(ylabel='', xlabel='')
    
    ax.set_title("IC Monthly Mean".format(period))
    
    return ax


# -----------------------------------------------------------------------------------
# Functions to Plot Others
def plot_event_bar_OLD(mean, std, ax):
    idx = mean.index
    
    DECIMAL_TO_PERCENT = 100.0
    ax.errorbar(idx, mean * DECIMAL_TO_PERCENT, yerr=std * DECIMAL_TO_PERCENT,
                marker='o',
                ecolor='lightblue', elinewidth=5)
    
    ax.set(xlabel='Period Length (trade days)', ylabel='Return (%)',
           title="Annual Return Mean and StdDev")
    ax.set(xticks=idx)
    return ax


def plot_event_bar(df, x, y, hue, ax):
    DECIMAL_TO_PERCENT = 100.0
    
    n = len(np.unique(df[hue]))
    palette_gen = (c for c in sns.color_palette("Reds_r", n))
    
    gp = df.groupby(hue)
    
    for p, dfgp in gp:
        idx = dfgp[x]
        mean = dfgp[y]
        # std = dfgp['Annu. Vol.']
        c = next(palette_gen)
        
        ax.errorbar(idx, mean * DECIMAL_TO_PERCENT,
                    marker='o', color=c,
                    # yerr=std * DECIMAL_TO_PERCENT, ecolor='lightblue', elinewidth=5,
                    label="{}".format(p))
    ax.axhline(0.0, color='k', ls='--', lw=1, alpha=.5)
    ax.set(xlabel='Period Length (trade days)', ylabel='Return (%)',
           title="Average Annual Return")
    ax.legend(loc='upper right')
    ax.set(xticks=idx)
    return ax


def plot_event_dist(df_events, date, axs):
    i = 0
    for period, ser in df_events.iteritems():
        ax = axs[i]
        sns.distplot(ser, ax=ax)
        ax.axvline(ser.mean(), lw=1, ls='--', label='Average', color='red')
        ax.legend(loc='upper left')
        ax.set(xlabel='Return (%)', ylabel='',
               title="{} Distribution of return after {:d} trade dats".format(date, period))
        # self.show_fig(fig, 'event_return_{:d}days.png'.format(my_period))
        i += 1
    
    # print(mean)


'''
def plot_event_dist_NEW(df_events, axs, grouper=None):
    i = 0
    def _plot(ser):
        ax = axs[i]
        sns.distplot(ser, ax=ax)
        ax.axvline(ser.mean(), lw=1, ls='--', label='Average', color='red')
        ax.legend(loc='upper left')
        ax.set(xlabel='Return (%)', ylabel='',
               title="Distribution of return after {:d} trade dats".format(period))
    if grouper is None:
    
    for (date, period), row in df_events.iterrows():
        ax = axs[i]
        sns.distplot(ser, ax=ax)
        ax.axvline(ser.mean(), lw=1, ls='--', label='Average', color='red')
        ax.legend(loc='upper left')
        ax.set(xlabel='Return (%)', ylabel='',
               title="Distribution of return after {:d} trade dats".format(period))
        # self.show_fig(fig, 'event_return_{:d}days.png'.format(my_period))
        i += 1
        
        # print(mean)

'''
def plot_calendar_distribution(signal, monthly_signal, yearly_signal, ax1, ax2):
    idx = signal.index.values
    start = jutil.convert_int_to_datetime(idx[0]).date()
    end = jutil.convert_int_to_datetime(idx[-1]).date()
    count = np.sum(yearly_signal.values.flatten())

    print("\n       " + "Calendar Distribution    ({} occurance from {} to {}):".format(count, start, end))

    # fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(16, 12), dpi=72)

    # sns.barplot(data=monthly_signal.reset_index(), x='Month', y='Time', ax=ax1£©
    # sns.barplot(x=monthly_signal.index.values, y=monthly_signal.values, ax=ax1)
    ax1.bar(monthly_signal.index, monthly_signal['Time'].values)
    ax1.axhline(monthly_signal.values.mean(), lw=1, ls='--', color='red', label='Average')
    ax1.legend(loc='upper right')
    months_str = ['Jan', 'Feb', 'March', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
    ax1.set(xticks=range(len(months_str)), xticklabels=months_str,
            title="Monthly Distribution",
            xlabel='Month', ylabel='Time')

    # sns.barplot(data=yearly_signal.reset_index(), x='Year', y='Times', ax=ax2, color='forestgreen')
    ax2.bar(yearly_signal.index, yearly_signal['Time'].values)
    ax2.axhline(yearly_signal.values.mean(), lw=1, ls='--', color='red', label='Average')
    ax2.legend(loc='upper right')
    ax2.set(xticks=yearly_signal.index,
            title="Yearly Distribution",
            xlabel='Month', ylabel='Time')


def plot_event_pvalue(pv, ax):
    idx = pv.index
    v = pv.values
    ax.plot(idx, v, marker='D')
    
    ax.set(xlabel='Period Length (trade days)', ylabel='p-value',
           title="P Value of Test: Mean(return) == 0")
    ax.set(xticks=idx)
    return ax


def plot_ic_decay(df_ic, ax):
    df_ic.mul(DECIMAL_TO_PCT).plot(marker='x', lw=1.2, ax=ax, cmap=COLOR_MAP)
    ax.axhline(0.0, color='k', ls='--', lw=0.7, alpha=.5)
    ax.set(xlabel="Period Length (trade days)", ylabel="IC (%)",
           title="IC Decay",
           xticks=df_ic.index,
           xlim=(0, df_ic.index[-1] + 1))


def plot_quantile_return_mean_std(dic, ax):
    n_quantiles = len(dic)
    palette_gen = (COLOR_MAP(x) for x in np.linspace(0, 1, n_quantiles))
    #palette_gen_light = (COLOR_MAP(x) for x in np.linspace(0, 1, n_quantiles))
    # palette_gen = (c for c in sns.color_palette("RdBu", n_quantiles, desat=0.5))
    # palette_gen =\
       # (c for c in sns.cubehelix_palette(n_quantiles,
       #                                             start=0, rot=0.5,
       #                                             dark=0.1, light=0.8, reverse=True,
       #                                             gamma=.9))
    # palette_gen_light = (c for c in sns.color_palette("RdBu", n_quantiles, desat=0.5))
    # palette_gen_light = (c for c in sns.cubehelix_palette(n_quantiles,
    #                                                start=0, rot=0.5,
    #                                                dark=0.1, light=0.8, reverse=True,
    #                                                gamma=.3))
    df_tmp = list(dic.values())[0]
    idx = df_tmp.columns
    offsets = np.linspace(-0.3, 0.3, n_quantiles)
    
    for i, (quantile, df) in enumerate(dic.items()):
        mean = df.loc['mean', :]
        std = df.loc['std', :]
        c = next(palette_gen)
        c_light = list(c)
        c_light[3] = c_light[3] * .2
        # c_light = next(palette_gen_light)
    
        ax.errorbar(idx + offsets[i], mean * DECIMAL_TO_PCT,
                    marker='x', color=c, lw=1.2,
                    yerr=std * DECIMAL_TO_PCT, ecolor=c_light, elinewidth=1,
                    label="Quantile {}".format(int(quantile)))
    ax.axhline(0.0, color='k', ls='--', lw=0.7, alpha=.5)
    ax.set(xlabel='Period Length (trade days)', ylabel='Return (%)',
           title="Mean & Std of Return",
           xticks=idx)
    ax.legend(loc='upper left')
    #ax.set(xticks=idx)


def plot_batch_backtest(df, ax):
    """
    
    Parameters
    ----------
    df : pd.DataFrame
    ax : axes

    """
    df = df.copy()
    df.index = jutil.convert_int_to_datetime(df.index)
    df.mul(DECIMAL_TO_PCT).plot(# marker='x',
                                lw=1.2, ax=ax, cmap=COLOR_MAP)
    ax.axhline(0.0, color='k', ls='--', lw=0.7, alpha=.5)
    ax.set(xlabel="Date", ylabel="Cumulative Return (%)",
           title="Cumulative Return for Different Buy Condition", )
    

